package de.lmu.ifi.pms.cep.dura.compiler.dura.desugaring;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.Stack;

import org.antlr.runtime.tree.CommonTree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraParser;


/**
 *  Why do we need Schema matching and why two of them?
 * 
 *  Originally because DuraC didn't (and maybe still doesn't) support sub term matching!
 *  
 * 
 * 
 *  It is now also a workaround for the fact that we don't have a proper semantic analysis phase yet.
 *  This would produce tables or annotated tree nodes containing type information.
 * 
 *  Since we don't have that, we need to improvise and pre-process the terms  in the program for
 *  the methods getType or determineType (or varieties thereof).
 *  
 *  Once there IS a proper semantic analysis phase and type information we can adapt these methods (and their usage)
 *  accordingly.
 *  
 *  Ex. Term:   a{ t{ var X } }
 *      Schema: a{ t{ q{ int }, r { int } } }
 *      
 *      What's the "type" of var X?  ->  It's the sub term sequence [q{ int }, r { int }]
 *      
 *      In this transformation we pre-process terms so that they take such an expanded form:
 *      
 *      a{ t{ q{ var 0_X }, r { var 1_X } } }
 *      
 *      
 *      
 *  We need two passes because during some transformations new rules are generated that need to be handled again.
 *  This might be movable to a dedicated per-rule handling method... but not now. 
 *  
 */
public class SchemaMatching extends Transformation {
	// We save which rules have been already handled
	public static IdentityHashMap<CommonTree, Boolean> handledRules = new IdentityHashMap<CommonTree, Boolean>();
	
	@Override
	public void apply(CommonTree tree) {
		for (CommonTree eventSpecification : collectEventSpecificationNodes(tree)) {
			if (handledRules.get(eventSpecification) != null) {
				continue;
			}
			
			CommonTree conjunctionOrDisjunction = (CommonTree) eventSpecification.getChild(1).getChild(0);
			
			HashMap<String, List<CommonTree>> varToTermSeqMap = new HashMap<String, List<CommonTree>>();
			HashMap<String, List<CommonTree>> varToSchemaTermSeqMap = new HashMap<String, List<CommonTree>>();
			
			ArrayList<CommonTree> terms = new ArrayList<CommonTree>();
			
			for (CommonTree query : collectAtomicQueries(conjunctionOrDisjunction)) {
				CommonTree term = (CommonTree) query.getChild(1).getChild(0);
				if (term.getType() == DuraParser.NEGATION) {
					term = (CommonTree) term.getChild(0);
				}
				
				HashMap<String, CommonTree> schemataMap = eventSchemataMap;
				if (query.getType() == DuraParser.STATE_QUERY) {
					schemataMap = stateSchemataMap;
				}
				
				CommonTree schema = schemataMap.get(term.getText());
				
				matchTermToSchemaTerm(term, schema, varToTermSeqMap, varToSchemaTermSeqMap);
				
				terms.add(term);
			}
			
			CommonTree headTerm = (CommonTree) eventSpecification.getChild(0).getChild(0);
			CommonTree headTermSchema = eventSchemataMap.get(headTerm.getText());
			matchTermToSchemaTerm(headTerm, headTermSchema, varToTermSeqMap, varToSchemaTermSeqMap);
			terms.add(headTerm);
			
			resolveSchemaMatchedVariables(terms, varToTermSeqMap);
			
			resetNextNameId();
			
			handledRules.put(eventSpecification, true);
		}
		
		
		// We handle reactive rules only in the second pass/schema matching
		for (CommonTree reactiveRule : collectReactiveRuleNodes(tree)) {
			if (handledRules.get(reactiveRule) != null) {
				continue;
			}
			
			// Handling the query part
			CommonTree body = (CommonTree) reactiveRule.getChild(0);
			CommonTree bodyComposition = (CommonTree) body.getChild(0);
			
			HashMap<String, List<CommonTree>> varToTermSeqMap = new HashMap<String, List<CommonTree>>();
			HashMap<String, List<CommonTree>> varToSchemaTermSeqMap = new HashMap<String, List<CommonTree>>();
			
			ArrayList<CommonTree> terms = new ArrayList<CommonTree>();
			
			for (CommonTree query : collectAtomicQueries(bodyComposition)) {
				CommonTree term = (CommonTree) query.getChild(1).getChild(0);
				if (term.getType() == DuraParser.NEGATION) {
					term = (CommonTree) term.getChild(0);
				}
				
				HashMap<String, CommonTree> schemataMap = eventSchemataMap;
				if (query.getType() == DuraParser.STATE_QUERY) {
					schemataMap = stateSchemataMap;
				}
				
				CommonTree schema = schemataMap.get(term.getText());
				
				matchTermToSchemaTerm(term, schema, varToTermSeqMap, varToSchemaTermSeqMap);
				
				terms.add(term);
			}
			
			
			// Handling the execution part
			CommonTree head = (CommonTree) reactiveRule.getChild(1);
			CommonTree concurrent = (CommonTree) head.getChild(0);
			
			for (CommonTree actionEx : collectAtomicActionExecutions(concurrent)) {
				CommonTree term;
				
				// DuraC actions don't have an identifier node, but Dura ones does.
				if (actionEx.getChildCount() > 1) {
					term = (CommonTree) actionEx.getChild(1);
				} else {
					term = (CommonTree) actionEx.getChild(0);
				}
				
				HashMap<String, CommonTree> schemataMap = actionSchemataMap;

				CommonTree schema = schemataMap.get(term.getText());
				
				matchTermToSchemaTerm(term, schema, varToTermSeqMap, varToSchemaTermSeqMap);
				
				terms.add(term);
			}
			
			
			resolveSchemaMatchedVariables(terms, varToTermSeqMap);
			
			
			resetNextNameId();
			
			handledRules.put(reactiveRule, true);
		}
		
		// TODO: State specifications
	}

	private LinkedList<CommonTree> collectVariablesInTerm(CommonTree tree) {
			LinkedList<CommonTree> nodes = new LinkedList<CommonTree>();
		Queue<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		nodeQueue.add(tree);
			
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();			
			
			if (node.getType() == DuraParser.VARIABLE) {
				nodes.add(node);
				continue;
			}
			
			if (node.getChildCount() > 0) {
				for (Object c : node.getChildren()) {
					CommonTree child = (CommonTree) c;
					nodeQueue.add(child);
				}
			}
		}
		return nodes;
	}
	
	@SuppressWarnings(value = "unchecked")
	private void resolveSchemaMatchedVariables(List<CommonTree> terms, HashMap<String, List<CommonTree>> varToTermSeqMap) {
		for (CommonTree term : terms) {
			for (CommonTree variable : collectVariablesInTerm(term)) {
				List<CommonTree> resolvedList = varToTermSeqMap.get(variable.getChild(0).getText());
				
				if (resolvedList != null) {
					int originalIndex = variable.getChildIndex();
					int index = originalIndex;
					
					CommonTree variableParent = (CommonTree) variable.getParent();
					
					List<CommonTree> variableSiblings = variableParent.getChildren();
					
					for (CommonTree resolved : resolvedList) {
						CommonTree resolvedCopy = copyTree(resolved);
						
						variableSiblings.add(index, resolvedCopy);
						index++;
					}
					
					variableParent.deleteChild(index);
					// The following line is not actually necessary since deleteChild() already calls this method.
					// http://www.antlr.org/api/Java/_base_tree_8java-source.html#l00154
					variableParent.freshenParentAndChildIndexes(originalIndex);
				}
			}
		}
	}

	/*
	 * Requires normalized terms.
	 */
	@SuppressWarnings(value = "unchecked")
	private boolean termSequencesEqual(List<CommonTree> termSeqA, List<CommonTree> termSeqB) {
		class TermEquation {
			public CommonTree left;
			public CommonTree right;
			
			TermEquation(CommonTree left, CommonTree right) {
				this.left = left;
				this.right = right;
			}
		}
		
		if (termSeqA.size() != termSeqB.size()) {
			return false;
		}
		
		Queue<TermEquation> termEqQueue = new LinkedList<TermEquation>();
		
		Iterator<CommonTree> termSeqAIter = termSeqA.iterator();
		Iterator<CommonTree> termSeqBIter = termSeqB.iterator();
		while (termSeqAIter.hasNext()) {
			CommonTree termA = termSeqAIter.next();
			
			if (termSeqBIter.hasNext()) {
				CommonTree termB = termSeqBIter.next();
				
				termEqQueue.add(new TermEquation(termA, termB));
			} else {
				return false;
			}
		}
		
		while(!termEqQueue.isEmpty()) {
			TermEquation equation = termEqQueue.remove();
			
			CommonTree left = equation.left;
			int leftType = left.getType();
			String leftText = left.getText();
			
			CommonTree right = equation.right;
			int rightType = right.getType();
			String rightText = right.getText();
			
			if (leftType != rightType) {
				return false;
			}
			if (!leftText.equals(rightText)) {
				return false;
			}
			if (left.getChildCount() != right.getChildCount()) {
				return false;
			}
			
			if (left.getChildCount() > 0) {
				int leftChildType = left.getChild(0).getType();
				int rightChildType = right.getChild(0).getType();
				
				if (rightChildType != leftChildType) {
					return false;
				}
				
				Iterator<CommonTree> leftChildrenIter = left.getChildren().iterator();
				Iterator<CommonTree> rightChildrenIter = right.getChildren().iterator();
				while (leftChildrenIter.hasNext()) {
					CommonTree childLeft = leftChildrenIter.next();
					
					if (rightChildrenIter.hasNext()) {
						CommonTree childRight = rightChildrenIter.next();
						
						termEqQueue.add(new TermEquation(childLeft, childRight));
					} else {
						return false;
					}
				}
			}
		}
		
		return true;
	}
	
	/*
	 * Requires normalized terms.
	 */
	@SuppressWarnings(value = "unchecked")
	private void matchTermToSchemaTerm(CommonTree termNode, CommonTree schemaTermNode,
			HashMap<String,	List<CommonTree>> varToTermSeqMap,
			HashMap<String, List<CommonTree>> varToSchemaTermSeqMap) {
		class TermSchemaTermEquation {
			public CommonTree term;
			public CommonTree schemaTerm;
			
			TermSchemaTermEquation (CommonTree term, CommonTree schema) {
				this.term = term;
				this.schemaTerm = schema;
			}
		}
		
		Queue<TermSchemaTermEquation> termSchemaTermEqQueue = new LinkedList<TermSchemaTermEquation>();
		termSchemaTermEqQueue.add(new TermSchemaTermEquation(termNode, schemaTermNode));
		
		while(!termSchemaTermEqQueue.isEmpty()) {
			TermSchemaTermEquation equation = termSchemaTermEqQueue.remove();
			
			CommonTree term = equation.term;
			int termType = term.getType();
			String termText = term.getText();
			
			CommonTree schemaTerm = equation.schemaTerm;
			int schemaTermType = schemaTerm.getType();
			String schemaTermText = schemaTerm.getText();
			
			if (termType != schemaTermType) {
				throw new RuntimeException("Schema mismatch!");
			}
			if (!termText.equals(schemaTermText)) {
				throw new RuntimeException("Schema mismatch!");
			}
			
			if (term.getChildCount() > 0) {
				CommonTree termChild = (CommonTree) term.getChild(0);
				int termChildType = termChild.getType();
				int schemaTermChildType = schemaTerm.getChild(0).getType();
				
				if (termChildType == DuraParser.VARIABLE) {
					String varId = termChild.getChild(0).getText();
					if (schemaTermChildType == DuraParser.TERM) {
						CommonTree schemaTermCopy = copyTree(schemaTerm);
					
						for (CommonTree basicType : collectBasicTypes(schemaTermCopy)) {
							CommonTree variable = createVariableNode(varId);
						
							int index = basicType.getChildIndex();
							basicType.getParent().replaceChildren(index, index, variable);
						}
						
						if (varToTermSeqMap.get(varId) == null) {
							varToTermSeqMap.put(varId, schemaTermCopy.getChildren());
						}
					}
					
					// The following actually performs basic type checking... but might be slow and all...
					// NOTE: This is only based on schemata and variables in terms. Arithmetic expressions etc.
					//       are not type-checked.
					List<CommonTree> temp = varToSchemaTermSeqMap.get(varId);
					if (temp == null) {
						varToSchemaTermSeqMap.put(varId, schemaTerm.getChildren());
					} else {
						if (!termSequencesEqual(temp, schemaTerm.getChildren())) {
							StringBuilder sb = new StringBuilder();
							sb.append("Schema for variable \"var " + varId + "\" ([");
							if (!temp.isEmpty()) {
								int i;
								for (i = 0; i < temp.size() - 1; i++) {
									sb.append(temp.get(i).toStringTree());
									sb.append(", ");
								}
								sb.append(temp.get(i).toStringTree());
							}
							sb.append("]) is not compatible with schema [");
							temp = schemaTerm.getChildren();
							if (!temp.isEmpty()) {
								int i;
								for (i = 0; i < temp.size() - 1; i++) {
									sb.append(temp.get(i).toStringTree());
									sb.append(", ");
								}
								sb.append(temp.get(i).toStringTree());
							}
							sb.append("].");
							
							throw new RuntimeException(sb.toString());
						}
					}
				
				} else if (termChildType == DuraParser.TERM) {
					if (schemaTermChildType != termChildType) {
						throw new RuntimeException("Schema mismatch!");
					}
					
					Iterator<CommonTree> termChildrenIter = term.getChildren().iterator();
					Iterator<CommonTree> schemaTermChildrenIter = schemaTerm.getChildren().iterator();
					while (termChildrenIter.hasNext()) {
						termChild = termChildrenIter.next();
						
						boolean labelMatched = false;
						
						while (schemaTermChildrenIter.hasNext()) {
							CommonTree schemaTermChild = schemaTermChildrenIter.next();
							
							// Omissions are fine, i.e. queries do not need to contain all schema terms to be valid.
							if (termChild.getText().equals(schemaTermChild.getText())) {
								termSchemaTermEqQueue.add(new TermSchemaTermEquation(termChild, schemaTermChild));
								labelMatched = true;
								break;
							}
						}
						
						if (!labelMatched) {
							throw new RuntimeException("Schema mismatch!");
						}
					}
				}
			}
		}
	}
	
	private LinkedList<CommonTree> collectBasicTypes(CommonTree tree) {
		LinkedList<CommonTree> nodes = new LinkedList<CommonTree>();
		Stack<CommonTree> nodeStack = new Stack<CommonTree>();
		nodeStack.add(tree);
		
		while(!nodeStack.isEmpty()) {
			CommonTree node = (CommonTree) nodeStack.pop();			
			
			if (node.getType() == DuraParser.BASIC_TYPE) {
				nodes.add(node);
				continue;
			}
			
			for (int i = node.getChildCount() - 1; i >= 0; i--) {
				CommonTree child = (CommonTree) node.getChild(i);
				nodeStack.push(child);
			}
		}
		return nodes;
	}
	
	private List<CommonTree> collectAtomicQueries(CommonTree conjunctionOrDisjunction) {
		ArrayList<CommonTree> nodes = new ArrayList<CommonTree>();
		
		assert(conjunctionOrDisjunction.getType() == DuraParser.CONJUNCTION || conjunctionOrDisjunction.getType() == DuraParser.DISJUNCTION);
		
		/*
		if (conjunctionOrDisjunction.getType() != DuraParser.CONJUNCTION && conjunctionOrDisjunction.getType() != DuraParser.DISJUNCTION) {
			return nodes;
		}
		*/
		
		Queue<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		nodeQueue.add(conjunctionOrDisjunction);
		
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();			
			
			switch(node.getType()) {
			case DuraParser.EVENT_QUERY:
			case DuraParser.STATE_QUERY:
				CommonTree q = (CommonTree) node.getChild(1); 
				if (q.getChild(q.getChildCount() - 1).getType() == DuraParser.TERM) {
					nodes.add(node);
				}
			}
			
			if (node.getChildCount() > 0) {
				for (Object c : node.getChildren()) {
					CommonTree child = (CommonTree) c;
					
					if (child.getType() != DuraParser.SUPPLEMENT) {
						nodeQueue.add(child);
					}
				}
			}
		}
		return nodes;
	}
	
	private List<CommonTree> collectAtomicActionExecutions(CommonTree concurrent) {
		ArrayList<CommonTree> nodes = new ArrayList<CommonTree>();
		
		assert(concurrent.getType() == DuraParser.CONCURRENT);
			
		Queue<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		nodeQueue.add(concurrent);
		
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();			
			
			if (node.getType() == DuraParser.ACTION_EXECUTION) {
				// DuraC actions don't have an identifier node, but Dura ones does.
				if (node.getChildCount() == 1 || node.getChild(1).getType() != DuraParser.CONCURRENT) {
					nodes.add(node);
				}
			}
			
			if (node.getChildCount() > 0) {
				for (Object c : node.getChildren()) {
					CommonTree child = (CommonTree) c;
					
					if (child.getType() != DuraParser.EXECUTION_CONSTRAINTS &&
						child.getType() != DuraParser.SUCCEEDS_ON &&
						child.getType() != DuraParser.FAILS_ON) {
						nodeQueue.add(child);
					}
				}
			}
		}
		return nodes;		
	}
}
