package de.lmu.ifi.pms.cep.dura.compiler.dura.desugaring;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

import org.antlr.runtime.CommonToken;
import org.antlr.runtime.tree.CommonTree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraParser;


/**
 * This transformation removes those expressions from the head term that are not allowed in the DURA
 * core language such as arithmetic expressions, interval operations (begin(), end()), durations.
 * 
 * Such expressions are replaced by variables and those variables are unified with the respective
 * expressions in a let statement in the body.
 * 
 * If a let statement already exists in the body the unifications will be added to that, if not 
 * a new let statement will be used.
 * 
 * @author scherr
 *
 */
public class SpecialExpressionsInHeadTerm extends Transformation {
	@Override
	public void apply(CommonTree tree) {
		for (CommonTree eventSpecification : collectEventSpecificationNodes(tree)) {
			String tempTreeString = eventSpecification.toStringTree();
			
			
			CommonTree head = (CommonTree) eventSpecification.getChild(0);
			
			assert(head.getType() == DuraParser.HEAD);
			
			CommonTree headTerm = (CommonTree) head.getChild(0);			
			
			ArrayList<CommonTree> unifications = new ArrayList<CommonTree>();
			for (CommonTree exp : collectSpecialExpressionNodes(headTerm)) {
				CommonTree variable = createVariableNode();
					
				int index = exp.getChildIndex();
				exp.getParent().replaceChildren(index, index, variable);
				
				CommonTree variableCopy = copyTree(variable);
				
				CommonTree schemaTerm = (CommonTree) eventSpecification.getParent().getChild(1).getChild(0);
				List<CommonTree> typeNodes = getTypeNodes(schemaTerm, exp);
				
				if (typeNodes.size() > 1 || typeNodes.get(0).getType() == DuraParser.TERM) {
					throw new RuntimeException("Placeholder or unification variable for expression " + exp.toStringTree() + " is of complex type!" );
				}
				
				CommonTree typeNode = typeNodes.get(0);
				
				CommonTree unification = createUnificationNode((CommonToken) typeNode.getToken(), variableCopy, exp);
				unifications.add(unification);
			}
			
			if (unifications.size() > 0) {
				CommonTree conjunctionOrDisjunction = (CommonTree) eventSpecification.getChild(1).getChild(0);
				
				CommonTree temp = (CommonTree) conjunctionOrDisjunction.getChild(conjunctionOrDisjunction.getChildCount() - 1);
				if (temp.getType() != DuraParser.SUPPLEMENT) {
					CommonTree supplement = new CommonTree(new CommonToken(DuraParser.SUPPLEMENT));
					conjunctionOrDisjunction.addChild(supplement);
					temp = supplement;
				}
							
				CommonTree maybeLet = null;
				int childCount = temp.getChildCount();
				if (childCount > 0) {
					maybeLet = (CommonTree) temp.getChild(childCount - 1);
				}
				CommonTree let;
				if (maybeLet != null && maybeLet.getType() == DuraParser.LET) {
					let = maybeLet;
				} else {
					let = new CommonTree(new CommonToken(DuraParser.LET));
					temp.addChild(let);
				}
				
				for (CommonTree unification : unifications) {
					let.addChild(unification);
				}
				
				if (EMIT_ANNOTATIONS) {
					CommonTree annotation = new CommonTree(new CommonToken(DuraParser.ANNOTATION, "ANNOTATION"));
					annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Result of transformation " + this.getClass().getSimpleName() + ".\"")));
					annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Handled expressions in head of deductive rule " + tempTreeString + ".\"")));
					eventSpecification.getParent().addChild(annotation);
				}
			}
			
			resetNextNameId();
		}
	}
	
	/**
	 * This method collects special expressions by breadth first search.
	 * Special expressions are:
	 * ^(ARITHMETIC_OP[...] exp exp) 
	 * ^(DURATION ^(timeUnit NUMBER) time*)
	 * ^(INTERVAL_OP[...] timeInterval)
	 * 
	 * Variables, numbers and strings are NOT special expressions.
	 * 
	 * This method is to be called mainly on (head) terms.
	 * 
	 * @param tree
	 * @return
	 */
	private List<CommonTree> collectSpecialExpressionNodes(CommonTree tree) {
		ArrayList<CommonTree> nodes = new ArrayList<CommonTree>();
		Queue<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		nodeQueue.add(tree);
		
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();			
			
			switch (node.getType()) {
			case DuraParser.ARITHMETIC_OP:
			case DuraParser.DURATION:
			case DuraParser.INTERVAL_OP:
				nodes.add(node);
				continue;
			}
			
			if (node.getChildCount() > 0) {
				for (Object c : node.getChildren()) {
					CommonTree child = (CommonTree) c;
					nodeQueue.add(child);
				}
			}
		}
		return nodes;
	}
}
