package de.lmu.ifi.pms.cep.dura.compiler.dura.desugaring;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.Stack;

import org.antlr.runtime.CommonToken;
import org.antlr.runtime.tree.CommonTree;
import org.antlr.runtime.tree.Tree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraParser;


/**
 * General, abstract transformation class also providing some utility methods.
 * 
 * @author scherr
 *
 */
public abstract class Transformation {
	protected static boolean EMIT_ANNOTATIONS = false;
	
	
	protected static int nextTransformationId = 0;
	protected int transformationId;
	protected int nextNameId;
	
	
	protected ArrayList<CommonTree> implicitTypeDefinitions = new ArrayList<CommonTree>();
	protected String[] implicitTypeDefinitionStrings = new String[] {
			"TYPE time-interval IS { begin{ timestamp }, end{ timestamp } } END"
	};
	
	protected static HashMap<String, CommonTree> typeDefinitionsMap = new HashMap<String, CommonTree>();
	
	
	
	
	protected static ArrayList<CommonTree> implicitEventSchemaTerms = new ArrayList<CommonTree>();
	protected static String[] implicitEventSchemaTermStrings = new String[] {
			"id{ identifier }",
			"reception-time{ begin{ timestamp }, end{ timestamp } }"
	};
	
	protected static ArrayList<CommonTree> implicitStateSchemaTerms = new ArrayList<CommonTree>();
	protected static String[] implicitStateSchemaTermString = new String[] {
			"id{ identifier }",
			"valid-time{ begin{ timestamp }, end{ timestamp } }"
	};
	
	protected static ArrayList<CommonTree> implicitActionEntailedEventSchemaTerms = new ArrayList<CommonTree>();
	protected static String[] implicitActionEntailedEventSchemaTermStrings = new String[] {
			"id{ identifier }",
			"reception-time{ begin{ timestamp }, end{ timestamp } }",
			"payload{ id{ identifier }, initiation-time{ begin{ timestamp } } }"
	};
	
	// These are generated/built in SchemaAnalysis but made available to all transformations in case
	// they need schema information.
	// Remember to update them in transformations that add/remove/modify schemata!
	protected static HashMap<String, CommonTree> explicitEventSchemataMap = null;
	protected static HashMap<String, CommonTree> eventSchemataMap = null;
	protected static HashMap<String, CommonTree> explicitStateSchemataMap = null;
	protected static HashMap<String, CommonTree> stateSchemataMap = null;
	protected static HashMap<String, CommonTree> explicitActionSchemataMap = null;
	protected static HashMap<String, CommonTree> actionSchemataMap = null;
	
	
	protected static CommonTree actionCompositionAugmentationTerm;
	protected static String compositionAugmentationTermString = "_composition { child-index{ integer }, parent-action-id{ identifier } } }";
	
	protected static HashSet<String> actionCompositionAugmentedActionSet = new HashSet<String>();
	
	
	protected Comparator<CommonTree> termComparator = new Comparator<CommonTree>() {
		@Override
		public int compare(CommonTree termA, CommonTree termB) {
			int temp = termA.getText().compareTo(termB.getText());
			
			if (temp == 0) {
				throw new RuntimeException("Term normalization issue! Labels must be unique per sequence.");
			}
			
			return temp;
		}
	};
	
	protected Transformation() {
		transformationId = nextTransformationId;
		nextTransformationId++;
	}
	
	/**
	 * This method applies the transformation on a given tree. The modifications are done in-place.
	 * 
	 * @param tree
	 */
	public abstract void apply(CommonTree tree);
	
	protected String createName(String hint) {
		String name = "_" + transformationId + "_" + nextNameId + (hint == null || hint.isEmpty() ? "" : "_" + hint);
		nextNameId++;
		
		return name;
	}
	
	protected String createName() {
		return createName("");
	}
	
	/**
	 * This method creates a new variable node with a locally unique identifier.
	 * If {@link #resetNextNameId()} is never called the identifier will be
	 * globally unique.
	 * 
	 * @return
	 */
	protected CommonTree createVariableNode() {
		CommonToken tempToken = new CommonToken(
				DuraParser.ID,
				createName());
		
		CommonTree tempTree = new CommonTree(new CommonToken(DuraParser.VARIABLE, "VARIABLE"));
		tempTree.addChild(new CommonTree(tempToken));
		return tempTree;
	}
	
	protected CommonTree createVariableNode(String hint) {
		CommonToken tempToken = new CommonToken(
				DuraParser.ID,
				createName(hint));
		
		CommonTree tempTree = new CommonTree(new CommonToken(DuraParser.VARIABLE, "VARIABLE"));
		tempTree.addChild(new CommonTree(tempToken));
		return tempTree;
	}
	
	/**
	 * This method resets the name identifier used by {@link #createVariableNode()}. It does not have to be
	 * called at all, but it is useful to keep names short and exploit the local scope.
	 * 
	 * It should only be called once for every local scope handled by a transformation.
	 */
	protected void resetNextNameId() {
		nextNameId = 0;
	}
	
	protected int getNextNameId() {
		return nextNameId;
	}
	
	protected void setNextNameId(int nextName) {
		this.nextNameId = nextName;
	}
	
	/**
	 * This method creates a unification node like: ^(UNIFICATION typedVariable with)
	 * No node type checking is performed here!
	 * 
	 * All parameters must be independent... (i.e. should not be children of anything) 
	 * 
	 * TODO: Node type checking?
	 * 
	 * @param type
	 * @param variable
	 * @param with
	 * @return
	 */
	@SuppressWarnings(value = "unchecked")
	protected CommonTree createUnificationNode(CommonToken type, CommonTree variable, CommonTree with) {
		assert(type.getType() == DuraParser.BASIC_TYPE || type.getType() == DuraParser.COMPOSITE_TYPE);
		assert(variable.getType() == DuraParser.VARIABLE);
		assert(variable.getParent() == null);
		assert(with.getParent() == null);
		
		// We have to transform the variable to a typed variable.
		variable.getToken().setType(DuraParser.TYPED_VARIABLE);
		variable.getToken().setText("TYPED_VARIABLE"); // optional
		variable.getChildren().add(0, new CommonTree(type));
		variable.freshenParentAndChildIndexes();
		
		CommonToken tempToken = new CommonToken(DuraParser.UNIFICATION);
		CommonTree tempTree = new CommonTree(tempToken);
		tempTree.addChild(variable);
		tempTree.addChild(with);
		
		return tempTree;
	}
	
	/**
	 * This method makes a deep copy of a tree.
	 * 
	 * TODO: Reevaluate whether the more simple and elegant, albeit possibly slower, recursive
	 *       implementation should rather be used.
	 * 
	 * @param tree
	 * @return
	 */
	protected CommonTree copyTree(CommonTree tree) {
		Queue<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		IdentityHashMap<CommonTree, CommonTree> copyMap = new IdentityHashMap<CommonTree, CommonTree>();
		
		nodeQueue.add(tree);
		
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();
			CommonTree nodeCopy = new CommonTree(new CommonToken(node.getToken()));
			
			copyMap.put(node, nodeCopy);
			
			CommonTree parentCopy = copyMap.get((CommonTree) node.getParent());
			if (parentCopy != null) {
				parentCopy.addChild(nodeCopy);
			}
			
			if (node.getChildCount() > 0) {
				for (Object c : node.getChildren()) {
					CommonTree child = (CommonTree) c;
					nodeQueue.add(child);
				}
			}
		}
		
		return copyMap.get(tree);
	}
	
	protected List<CommonTree> collectSchemata(CommonTree tree) {
		ArrayList<CommonTree> nodes = new ArrayList<CommonTree>();
		CommonTree searchRoot;
		Tree temp = tree.getChild(0).getChild(1);
		
		assert(temp.getType() == DuraParser.DEFINITIONS);
		
		// This should never happen because of $defs+
		if (temp.getChildCount() > 0) {
			searchRoot = (CommonTree) temp;
		} else {
			return nodes;
		}
		
		for (int i = 0; i < searchRoot.getChildCount(); i++) {
			temp = searchRoot.getChild(i);
			
			nodes.add((CommonTree) temp.getChild(1));
		}
		
		return nodes;
	}
	
	@SuppressWarnings(value = "unchecked")
	protected void normalizeTerm(CommonTree term) {
		assert (term.getType() == DuraParser.TERM);
		
		Queue<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		nodeQueue.add(term);
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();
			if (node.getChildCount() > 0) {
				// If a TERM does have a child that is not a TERM we stop for this branch.
				if (node.getChild(0).getType() != DuraParser.TERM) {
					continue;
				}
				
				Collections.sort(node.getChildren(), termComparator);
				
				node.freshenParentAndChildIndexes();
				
				for (Object c : node.getChildren()) {
					CommonTree child = (CommonTree) c;
					nodeQueue.add(child);
				}
			}
		}
	}
	
	protected List<CommonTree> collectTermNodes(CommonTree tree) {
		ArrayList<CommonTree> nodes = new ArrayList<CommonTree>();
		Queue<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		nodeQueue.add(tree);
		
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();
			
			switch (node.getType()) {
			case DuraParser.TERM:
				nodes.add(node);
				break;
				
			case DuraParser.ARITHMETIC_OP:
			case DuraParser.AGGREGATION_OP:
			case DuraParser.AGGREGATION:
			case DuraParser.SUPPLEMENT:
				continue;
			}
			
			if (node.getChildCount() > 0) {
				for (Object c : node.getChildren()) {
					CommonTree child = (CommonTree) c;
					nodeQueue.add(child);
				}
			}
		}
		
		return nodes;
	}
	
	/*
	 * Works also on general, non-normalized terms. 
	 * For normalized terms treesEqual can be used!!! 
	 */
	protected boolean termsEqual(CommonTree termA, CommonTree termB) {
		class TermEquation {
			public CommonTree left;
			public CommonTree right;
			
			TermEquation(CommonTree left, CommonTree right) {
				this.left = left;
				this.right = right;
			}
		}
		
		Queue<TermEquation> termEqQueue = new LinkedList<TermEquation>();
		
		HashMap<String, CommonTree> labelToTermMap = new HashMap<String, CommonTree>();
		
		termEqQueue.add(new TermEquation(termA, termB));
		
		while(!termEqQueue.isEmpty()) {
			TermEquation equation = termEqQueue.remove();
			
			CommonTree left = equation.left;
			int leftType = left.getType();
			String leftText = left.getText();
			
			CommonTree right = equation.right;
			int rightType = right.getType();
			String rightText = right.getText();
			
			if (leftType != rightType) {
				return false;
			}
			if (!leftText.equals(rightText)) {
				return false;
			}
			if (left.getChildCount() != right.getChildCount()) {
				return false;
			}
			
			if (left.getChildCount() > 0) {
				int leftChildType = left.getChild(0).getType();
				int rightChildType = right.getChild(0).getType();
				
				if (rightChildType != leftChildType) {
					return false;
				}
					
				labelToTermMap = new HashMap<String, CommonTree>();
				for (Object c : right.getChildren()) {
					CommonTree child = (CommonTree) c;
					String childText = child.getText();
						
					labelToTermMap.put(childText, child);
				}
					
				for (Object c : left.getChildren()) {
					CommonTree child = (CommonTree) c;
					
					termEqQueue.add(new TermEquation(child, labelToTermMap.get(child.getText())));
				}
			}
		}
		
		return true;
	}
	
	@SuppressWarnings(value = "unchecked")
	protected boolean treesEqual(CommonTree treeA, CommonTree treeB) {
		class TermEquation {
			public CommonTree left;
			public CommonTree right;
			
			TermEquation(CommonTree left, CommonTree right) {
				this.left = left;
				this.right = right;
			}
		}
		
		Queue<TermEquation> termEqQueue = new LinkedList<TermEquation>();
		
		termEqQueue.add(new TermEquation(treeA, treeB));
		
		while(!termEqQueue.isEmpty()) {
			TermEquation equation = termEqQueue.remove();
			
			CommonTree left = equation.left;
			int leftType = left.getType();
			String leftText = left.getText();
			
			CommonTree right = equation.right;
			int rightType = right.getType();
			String rightText = right.getText();
			
			if (leftType != rightType) {
				return false;
			}
			if (!leftText.equals(rightText)) {
				return false;
			}
			if (left.getChildCount() != right.getChildCount()) {
				return false;
			}
			
			if (left.getChildCount() > 0) {
				int leftChildType = left.getChild(0).getType();
				int rightChildType = right.getChild(0).getType();
				
				if (rightChildType != leftChildType) {
					return false;
				}
				
				Iterator<CommonTree> leftChildrenIter = left.getChildren().iterator();
				Iterator<CommonTree> rightChildrenIter = right.getChildren().iterator();
				while (leftChildrenIter.hasNext()) {
					CommonTree childLeft = leftChildrenIter.next();
					
					if (rightChildrenIter.hasNext()) {
						CommonTree childRight = rightChildrenIter.next();
						
						termEqQueue.add(new TermEquation(childLeft, childRight));
					} else {
						return false;
					}
				}
			}
		}
		
		return true;
	}
	
	/**
	 * This method collects event specification nodes 
	 * (^(DETECT ^(HEAD constructTerm (^(AGGREGATION ^(GROUPING binding+)))?) ^(BODY query)))
	 * by walking the known tree structure on two levels (two indexes, two loops).
	 * 
	 * @param tree
	 * @return
	 */
	protected List<CommonTree> collectEventSpecificationNodes(CommonTree tree) {
		ArrayList<CommonTree> nodes = new ArrayList<CommonTree>();
		CommonTree searchRoot;
		Tree temp = tree.getChild(0).getChild(1);
		
		assert(temp.getType() == DuraParser.DEFINITIONS);
		
		// This should never happen because of $defs+
		if (temp.getChildCount() > 0) {
			searchRoot = (CommonTree) temp;
		} else {
			return nodes;
		}
		
		for (int i = 0; i < searchRoot.getChildCount(); i++) {
			temp = searchRoot.getChild(i);
			
			if (temp.getType() == DuraParser.EVENT) {
				for (int j = 2; j < temp.getChildCount(); j++) {
					if (temp.getChild(j).getType() == DuraParser.DETECT) {
						nodes.add((CommonTree) temp.getChild(j));
					}
				}
			}
		}
		
		return nodes;
	}
	
	/**
	 * This method provides a general node collection and filtering mechanism by breadth first search.
	 * It is expected to be slower than specialized collector methods as it is agnostic of the Dura
	 * AST structure.  
	 * 
	 * If a node type occurs in both searchTypes and stopTypes only the respective first (or highest)
	 * occurrences of nodes of that type along a path of the AST will be collected. 
	 * 
	 * @param tree
	 * @param searchTypes	the node types to collect
	 * @param stopTypes 	the node types at which the search tree is cut
	 * @return
	 */
	protected List<CommonTree> collectNodes(CommonTree tree, int[] searchTypes, int[] stopTypes) {
		ArrayList<CommonTree> nodes = new ArrayList<CommonTree>();
		HashSet<Integer> nodeTypesSet = new HashSet<Integer>();
		if (searchTypes != null) {
			for (int nodeType : searchTypes) {
				nodeTypesSet.add(Integer.valueOf(nodeType));
			}
		}
		HashSet<Integer> stopTypesSet = new HashSet<Integer>();
		if (stopTypes != null) {
			for (int stopType : stopTypes) {
				stopTypesSet.add(Integer.valueOf(stopType));
			}
		}
		Queue<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		nodeQueue.add(tree);
		
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();
			
			if (nodeTypesSet.contains(node.getType())) {
				nodes.add(node);
			}
			
			if (stopTypesSet.contains(node.getType())) {
				continue;
			}
			
			if (node.getChildCount() > 0) {
				for (Object c : node.getChildren()) {
					CommonTree child = (CommonTree) c;
					nodeQueue.add(child);
				}
			}
		}
		
		return nodes;
	}
	
	protected List<CommonTree> getEnvironment(CommonTree query) {
		assert(query.getType() == DuraParser.NEGATION || query.getType() == DuraParser.EXISTS || query.getType() == DuraParser.EVENT_QUERY || query.getType() == DuraParser.STATE_QUERY);
		
		CommonTree composition = (CommonTree) query.getParent();
		
		ArrayList<CommonTree> nodes = new ArrayList<CommonTree>();
		
		int endIndex = composition.getChildCount() - 1;
		if (composition.getChild(endIndex).getType() == DuraParser.SUPPLEMENT) {
			endIndex--;
		}
		
		for (int i = 0; i <= endIndex; i++) {
			CommonTree q = (CommonTree) composition.getChild(i);
			if (q != query) {
				nodes.add(q);
			}
		}
		
		return nodes;
	}
	
	protected HashSet<CommonTreeSetWrapper> termBindings(CommonTree term) {
		assert(term.getType() == DuraParser.TERM);
		
		HashSet<CommonTreeSetWrapper> set = new HashSet<CommonTreeSetWrapper>();
		
		Queue<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		nodeQueue.add(term);
		
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();			
			
			if (node.getType() == DuraParser.IDENTIFIER || node.getType() == DuraParser.VARIABLE) {
				set.add(new CommonTreeSetWrapper(node));
			}
			
			if (node.getChildCount() > 0) {
				for (Object c : node.getChildren()) {
					CommonTree child = (CommonTree) c;
					nodeQueue.add(child);
				}
			}
		}
			
		return set;
	}
	
	protected CommonTree identifier(CommonTree query) {
		assert(query.getType() == DuraParser.NEGATION || query.getType() == DuraParser.EXISTS || query.getType() == DuraParser.EVENT_QUERY || query.getType() == DuraParser.STATE_QUERY);
		
		if (query.getType() == DuraParser.NEGATION || query.getType() == DuraParser.EXISTS) {
			query = (CommonTree) query.getChild(0);
		}
		
		return (CommonTree) query.getChild(0);
	}
	
	protected CommonTree getVariableFromUnification(CommonTree unification) {
		assert(unification.getType() == DuraParser.UNIFICATION);
		
		CommonTree variableCopy = copyTree((CommonTree) unification.getChild(0));
		variableCopy.getToken().setType(DuraParser.VARIABLE);
		variableCopy.getToken().setText("VARIABLE");
		variableCopy.deleteChild(0);
		
		return variableCopy;
	}
	
	protected HashSet<CommonTreeSetWrapper> suppDefs(CommonTree supplement) {
		assert(supplement.getType() == DuraParser.SUPPLEMENT);
		
		HashSet<CommonTreeSetWrapper> set = new HashSet<CommonTreeSetWrapper>();
		
		Queue<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		nodeQueue.add(supplement);
		
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();			
			
			if (node.getType() == DuraParser.UNIFICATION) {
				set.add(new CommonTreeSetWrapper(getVariableFromUnification(node)));
			}
			
			if (node.getChildCount() > 0) {
				for (Object c : node.getChildren()) {
					CommonTree child = (CommonTree) c;
					
					if (node.getType() != DuraParser.CONDITION) {
						nodeQueue.add(child);
					}
				}
			}
		}
			
		return set;
	}
	
	protected class CommonTreeSetWrapper {
		private CommonTree tree;
		
		public CommonTreeSetWrapper(CommonTree tree) {
			this.tree = tree;
		}
		
		public boolean equals(Object obj) {
			return treesEqual(((CommonTreeSetWrapper) obj).tree, tree);
		}
		
		public int hashCode() {
			return tree.toStringTree().hashCode();
		}
		
		public CommonTree getTree() {
			return tree;
		}
	}
	
	protected List<CommonTree> collectActionSpecificationNodes(CommonTree tree) {
		ArrayList<CommonTree> nodes = new ArrayList<CommonTree>();
		CommonTree searchRoot;
		Tree temp = tree.getChild(0).getChild(1);
		
		assert(temp.getType() == DuraParser.DEFINITIONS);
		
		// This should never happen because of $defs+
		if (temp.getChildCount() > 0) {
			searchRoot = (CommonTree) temp;
		} else {
			return nodes;
		}
		
		for (int i = 0; i < searchRoot.getChildCount(); i++) {
			temp = searchRoot.getChild(i);
			
			if (temp.getType() == DuraParser.ACTION) {		
				if (temp.getChildCount() > 4) {
					if (temp.getChild(4).getType() == DuraParser.FOR) {
						nodes.add((CommonTree) temp.getChild(4));
					}
				}
			}
		}
		
		return nodes;
	}
	
	protected List<CommonTree> collectReactiveRuleNodes(CommonTree tree) {
		ArrayList<CommonTree> nodes = new ArrayList<CommonTree>();
		CommonTree searchRoot;
		Tree temp = tree.getChild(0).getChild(1);
		
		assert(temp.getType() == DuraParser.DEFINITIONS);
		
		// This should never happen because of $defs+
		if (temp.getChildCount() > 0) {
			searchRoot = (CommonTree) temp;
		} else {
			return nodes;
		}
		
		for (int i = 0; i < searchRoot.getChildCount(); i++) {
			temp = searchRoot.getChild(i);
			
			if (temp.getType() == DuraParser.DO) {
				nodes.add((CommonTree) temp);
			}
		}
		
		return nodes;
	}
	
	protected void addImplicitActionEntailedEventSchemata(String actionSchemaTermText, CommonTree actionSchemaTerm) {
		// Semantic analysis has to guarantee that these events have not been defined by a programmer.
		// The are implicit by nature, i.e. they depend on what the action schema looks like.
		//
		// Note that all could share own schema... but maybe this will change in the future, so each event gets its
		// own schema copy. 
		
		CommonTree initiatedSchema = new CommonTree(new CommonToken(DuraParser.TERM, actionSchemaTermText + "$initiated"));
		CommonTree suceededSchema = new CommonTree(new CommonToken(DuraParser.TERM, actionSchemaTermText + "$succeeded"));
		CommonTree failedSchema = new CommonTree(new CommonToken(DuraParser.TERM, actionSchemaTermText + "$failed"));
				
		joinTermCopiesWithTermChildren(implicitActionEntailedEventSchemaTerms, initiatedSchema);
		joinTermCopiesWithTermChildren(implicitActionEntailedEventSchemaTerms, suceededSchema);
		joinTermCopiesWithTermChildren(implicitActionEntailedEventSchemaTerms, failedSchema);
		
		if (actionSchemaTerm.getChildCount() > 0) {
			ArrayList<CommonTree> payloadTerms = new ArrayList<CommonTree>(actionSchemaTerm.getChildren());
			
			int payloadIndex; 
			for (payloadIndex = 0; payloadIndex < implicitActionEntailedEventSchemaTerms.size(); payloadIndex++) {
				if (implicitActionEntailedEventSchemaTerms.get(payloadIndex).getText().equals("payload")) {
					break;
				}
			}
			
			joinTermCopiesWithTermChildren(payloadTerms, (CommonTree) initiatedSchema.getChild(payloadIndex));
			joinTermCopiesWithTermChildren(payloadTerms, (CommonTree) suceededSchema.getChild(payloadIndex));
			joinTermCopiesWithTermChildren(payloadTerms, (CommonTree) failedSchema.getChild(payloadIndex));
		}
		
		eventSchemataMap.put(initiatedSchema.getText(), initiatedSchema);
		eventSchemataMap.put(suceededSchema.getText(), suceededSchema);
		eventSchemataMap.put(failedSchema.getText(), failedSchema);
	}
	
	@SuppressWarnings(value="unchecked")
	protected void joinTermCopiesWithTermChildren(ArrayList<CommonTree> joinTerms, CommonTree term) {
		if (term.getChildCount() == 0) {
			for (CommonTree j : joinTerms) {
				term.addChild(copyTree(j));
			}
			
			return;
		}
		
		Iterator<CommonTree> termChildrenIter = term.getChildren().iterator();
		
		ArrayList<CommonTree> termSeq = new ArrayList<CommonTree>();
		
		int i = 0;
		while (termChildrenIter.hasNext()) {
			CommonTree childTerm = termChildrenIter.next();
			
			for (; i < joinTerms.size(); i++) {
				CommonTree joinTerm = joinTerms.get(i);
				
				int comp = childTerm.getText().compareTo(joinTerm.getText());
				if (comp == 0) {
					if (!treesEqual(childTerm, joinTerm)) {
						throw new RuntimeException("Cannot join (" + joinTerm.toStringTree() + ") with children of (" + term.toStringTree() + ").");
					}
				} else if (comp < 0) {
					break;
				} else {
					termSeq.add(copyTree(joinTerm));
				}
			}
			
			termSeq.add(childTerm);
		}
		for (; i < joinTerms.size(); i++) {
			termSeq.add(copyTree(joinTerms.get(i)));
		}

		List<CommonTree> termChildren = term.getChildren();
		
		termChildren.clear();
		termChildren.addAll(termSeq);
		term.freshenParentAndChildIndexes();
	}
	
	// TODO: Walk up down by name instead of child index.
	protected List<CommonTree> getTypeNodes(CommonTree schemaTerm, CommonTree termLeaf) {
		
		/*
		ArrayList<Integer> pathToTermRoot = new ArrayList<Integer>();
		CommonTree tempTerm = termLeaf;
		
		CommonTree parent = (CommonTree) tempTerm.getParent();
		while (parent.getType() == DuraParser.TERM) {
			pathToTermRoot.add(tempTerm.getChildIndex());
			tempTerm = parent;
			parent = (CommonTree) tempTerm.getParent();
		}
		
		tempTerm = schemaTerm;
		
		for (int i = pathToTermRoot.size() - 1; i >= 0; i--) {
			tempTerm = (CommonTree) tempTerm.getChild(pathToTermRoot.get(i));
		}
		
		if (tempTerm.getType() == DuraParser.BASIC_TYPE || tempTerm.getType() == DuraParser.COMPOSITE_TYPE) {
			return Collections.singletonList(copyTree(tempTerm));
		} else {
			throw new RuntimeException("Type matching failed!");
		}
		*/
		
		CommonTree tempTerm = termLeaf;
		
		ArrayList<String> pathToTermRoot = new ArrayList<String>();
		
		CommonTree parent = (CommonTree) tempTerm.getParent();
		while (parent.getType() == DuraParser.TERM) {
			pathToTermRoot.add(parent.getText());
			tempTerm = parent;
			parent = (CommonTree) tempTerm.getParent();
		}
		
		tempTerm = schemaTerm;
		
		
		// tempTerm is the schema term
		for (int i = pathToTermRoot.size() - 2; i >= 0; i--) {
			String label = pathToTermRoot.get(i);
			
			if (tempTerm.getChildCount() > 0) {
				boolean foundMatchingTermLabel = false; 
				
				for (Object c : tempTerm.getChildren()) {
					CommonTree schemaChild = (CommonTree) c;
					
					if (schemaChild.getType() != DuraParser.TERM) {
						if (schemaChild.getType() == DuraParser.COMPOSITE_TYPE) {
							CommonTree typeDef = typeDefinitionsMap.get(schemaChild.getText());
							
							for (int j = 0; j < typeDef.getChildCount() - 1; j++) {
								CommonTree t = (CommonTree) typeDef.getChild(j);
								
								if (t.getType() != DuraParser.TERM) {
									break;
								} else {
									if (label.equals(t.getText())) {
										foundMatchingTermLabel = true;
										
										tempTerm = t;
										break;
									}
								}
							}
						} else {
							break;
						}
					} else {
						String schemaChildLabel = schemaChild.getText();
						
						if (label.equals(schemaChildLabel)) {
							foundMatchingTermLabel = true;
							
							tempTerm = schemaChild;
							break;
						}
					}
				}
				
				if (!foundMatchingTermLabel) {
					throw new RuntimeException("Type matching failed!");
				}
			} else {
				break;
				// throw new RuntimeException("Type matching failed!");
			}
		}
		
		ArrayList<CommonTree> typeNodes = new ArrayList<CommonTree>();
		for (int i = 0; i < tempTerm.getChildCount(); i++) {
			typeNodes.add(copyTree((CommonTree) tempTerm.getChild(i)));
		}
		if (typeNodes.size() == 0) {
			throw new RuntimeException("Type matching failed!" );
		}
		
		return typeNodes;
	}
	
	protected List<CommonTree> determineAndGetTypeNodes(CommonTree composition, CommonTree variable) {
		assert(variable.getType() == DuraParser.VARIABLE);
		
		String variableName = variable.getChild(0).getText();
		
		Queue<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		nodeQueue.add(composition);
		
		CommonTree tempTerm = null;
		
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();			
			
			if (node.getType() == DuraParser.VARIABLE && node.getParent().getType() == DuraParser.TERM) {
				// We are only interested in untyped variables if they occur in query terms.
				
				if (node.getChild(0).getText().equals(variableName)) {
					tempTerm = node;
					break;
				}
			} else if (node.getType() == DuraParser.TYPED_VARIABLE) {
				if (node.getChild(1).getText().equals(variableName)) {
					return Collections.singletonList(copyTree((CommonTree) node.getChild(0)));
				}
			}
			
			if (node.getChildCount() > 0) {
				for (Object c : node.getChildren()) {
					CommonTree child = (CommonTree) c;
					nodeQueue.add(child);
				}
			}
		}
		
		if (tempTerm == null) {
			throw new RuntimeException("Type matching failed (variable: " + variableName + ")!");
		}
		
		ArrayList<String> pathToTermRoot = new ArrayList<String>();
		
		CommonTree parent = (CommonTree) tempTerm.getParent();
		while (parent.getType() == DuraParser.TERM) {
			pathToTermRoot.add(parent.getText());
			tempTerm = parent;
			parent = (CommonTree) tempTerm.getParent();
		}
		
		String termLabel = tempTerm.getText();
		
		int schemaType = (parent.getParent().getType() == DuraParser.EVENT_QUERY ? DuraParser.EVENT : DuraParser.STATEFUL_OBJECT);
		
		if (schemaType == DuraParser.EVENT) {
			tempTerm = eventSchemataMap.get(termLabel);
		} else {
			tempTerm = stateSchemataMap.get(termLabel);
		}
		
		if (tempTerm == null) {
			throw new RuntimeException("Type matching failed (variable: " + variableName + ")!");
		}
		
		
		// tempTerm is the schema term
		for (int i = pathToTermRoot.size() - 2; i >= 0; i--) {
			String label = pathToTermRoot.get(i);
			
			if (tempTerm.getChildCount() > 0) {
				boolean foundMatchingTermLabel = false; 
				
				for (Object c : tempTerm.getChildren()) {
					CommonTree schemaChild = (CommonTree) c;
					
					if (schemaChild.getType() != DuraParser.TERM) {
						if (schemaChild.getType() == DuraParser.COMPOSITE_TYPE) {
							CommonTree typeDef = typeDefinitionsMap.get(schemaChild.getText());
							
							for (int j = 0; j < typeDef.getChildCount() - 1; j++) {
								CommonTree t = (CommonTree) typeDef.getChild(j);
								
								if (t.getType() != DuraParser.TERM) {
									break;
								} else {
									if (label.equals(t.getText())) {
										foundMatchingTermLabel = true;
										
										tempTerm = t;
										break;
									}
								}
							}
						} else {
							break;
						}
					} else {
						String schemaChildLabel = schemaChild.getText();
						
						if (label.equals(schemaChildLabel)) {
							foundMatchingTermLabel = true;
							
							tempTerm = schemaChild;
							break;
						}
					}
				}
				
				if (!foundMatchingTermLabel) {
					throw new RuntimeException("Type matching failed (variable: " + variableName + ")!");
				}
			} else {
				break;
				// throw new RuntimeException("Type matching failed (variable: " + variableName + ")!");
			}
		}

		ArrayList<CommonTree> typeNodes = new ArrayList<CommonTree>();
		for (int i = 0; i < tempTerm.getChildCount(); i++) {
			typeNodes.add(copyTree((CommonTree) tempTerm.getChild(i)));
		}
		if (typeNodes.size() == 0) {
			throw new RuntimeException("Type matching failed (variable: " + variableName + ")!");
		}
		
		return typeNodes;
	}
	
	
	protected void addActionCompositionInformation(CommonTree tree, String actionName) {
		CommonTree actionSchemaTerm = actionSchemataMap.get(actionName);
		
		if (actionSchemaTerm == null) {	
			throw new RuntimeException("No action of name " + actionName + " has been defined!");
		}
		
		CommonTree explicitActionSchemaTerm = explicitActionSchemataMap.get(actionName);
		
		if (explicitActionSchemaTerm == null) {
			// Extension of external events that are not explicitly represented in the code?
			// Siehe E-Mail vom 11. Okt. 2011.
			
			Tree defs = tree.getChild(0).getChild(1);
			
			assert(defs.getType() == DuraParser.DEFINITIONS);
			
			CommonTree def = new CommonTree(new CommonToken(DuraParser.ACTION, actionName));
			def.addChild(new CommonTree(new CommonToken(DuraParser.MODIFIERS)));
			
			CommonTree schema = new CommonTree(new CommonToken(DuraParser.SCHEMA));
			
			explicitActionSchemaTerm = copyTree(actionSchemaTerm);
			augmentWithActionCompositionTerm(explicitActionSchemaTerm);
			
			schema.addChild(explicitActionSchemaTerm);
			schema.addChild(new CommonTree(new CommonToken(DuraParser.SUPPLEMENT)));
			
			def.addChild(schema);
			
			explicitActionSchemataMap.put(actionName, explicitActionSchemaTerm);
			
			// TODO: Has to be tested.
		}
		
		augmentWithActionCompositionTerm(explicitActionSchemaTerm);
		augmentWithActionCompositionTerm(actionSchemaTerm);
		
		CommonTree initiatedSchema = eventSchemataMap.get(actionName + "$initiated");
		CommonTree suceededSchema = eventSchemataMap.get(actionName + "$succeeded");
		CommonTree failedSchema = eventSchemataMap.get(actionName + "$failed");
		
		int payloadIndex; 
		for (payloadIndex = 0; payloadIndex < initiatedSchema.getChildCount(); payloadIndex++) {
			if (initiatedSchema.getChild(payloadIndex).getText().equals("payload")) {
				break;
			}
		}
		
		augmentWithActionCompositionTerm((CommonTree) initiatedSchema.getChild(payloadIndex));
		augmentWithActionCompositionTerm((CommonTree) suceededSchema.getChild(payloadIndex));
		augmentWithActionCompositionTerm((CommonTree) failedSchema.getChild(payloadIndex));
	}

	protected void augmentWithActionCompositionTerm(CommonTree term) {
		CommonTree addTerm = copyTree(actionCompositionAugmentationTerm);
		
		if (term.getChildCount() == 0) {
			term.addChild(addTerm);
		} else {
			int i;
			
			for (i = 0; i < term.getChildCount(); i++) {
				CommonTree childTerm = (CommonTree) term.getChild(i);
				
				int comp = childTerm.getText().compareTo(addTerm.getText());
				if (comp == 0) {
					throw new RuntimeException("This should not happen. Cannot join (" + addTerm.toStringTree() + ") with children of (" + term.toStringTree() + ").");
				} else if (comp > 0) {
					break;
				}
			}
			
			term.getChildren().add(i, addTerm);
			term.freshenParentAndChildIndexes();
		}
	}
	
	protected void replaceExpressionsInConcurrent(CommonTree concurrent, HashMap<String, CommonTree> varNameToExpressionMap, HashMap<String, List<CommonTree>> varNameToTypeNodesMap) {
		// Using a stack is important!! We need to save the last action name traversed...
		Stack<CommonTree> nodeStack = new Stack<CommonTree>();
		nodeStack.add(concurrent);
		
		String actionName = null;
		
		HashMap<String, String> eventIdentNameToVarNameMap = new HashMap<String, String>();
		HashMap<String, String> stateIdentNameToVarNameMap = new HashMap<String, String>();
		HashMap<String, String> actionIdentNameToVarNameMap = new HashMap<String, String>();
		
		
		while(!nodeStack.isEmpty()) {
			CommonTree node = (CommonTree) nodeStack.pop();			
			
			
			if (node.getParent().getType() == DuraParser.TERM) {
				CommonTree var;
				String varName;
				int index;
				CommonTree schemaTerm;
				List<CommonTree> typeNodes;
				
				switch (node.getType()) {	
				case DuraParser.ARITHMETIC_OP:
				case DuraParser.DURATION:
				case DuraParser.INTERVAL_OP:
					var = createVariableNode();
					varName = var.getChild(0).getText();
					
					index = node.getChildIndex();
					node.getParent().replaceChildren(index, index, var);
					
					varNameToExpressionMap.put(varName, node);
					
					schemaTerm = actionSchemataMap.get(actionName);
					typeNodes = getTypeNodes(schemaTerm, node);
					
					if (typeNodes.size() > 1 || typeNodes.get(0).getType() == DuraParser.TERM) {
						throw new RuntimeException("Placeholder or unification variable for expression " + node.toStringTree() + " is of complex type!" );
					}
									
					varNameToTypeNodesMap.put(varName, typeNodes);
					
					continue;
					// break;
				
				// Important!
				case DuraParser.IDENTIFIER:
					if (node.getChildCount() > 0) {
						String identName = node.getChild(0).getText();
						varName = null;
						
						if (node.getText().equals("event")) {
							varName = eventIdentNameToVarNameMap.get(identName);
						} else if (node.getText().equals("state")) {
							varName = stateIdentNameToVarNameMap.get(identName);
						} else if (node.getText().equals("action")) {
							varName = actionIdentNameToVarNameMap.get(identName);
						}
						
						
						if (varName == null) {
							var = createVariableNode();
							varName = var.getChild(0).getText();
							
							varNameToExpressionMap.put(varName, node);
							
							schemaTerm = actionSchemataMap.get(actionName);
							typeNodes = getTypeNodes(schemaTerm, node);
							
							if (typeNodes.size() > 1 || typeNodes.get(0).getType() == DuraParser.TERM) {
								throw new RuntimeException("Placeholder or unification variable for expression " + node.toStringTree() + " is of complex type!" );
							}
											
							varNameToTypeNodesMap.put(varName, typeNodes);
							
							
							if (node.getText().equals("event")) {
								eventIdentNameToVarNameMap.put(identName, varName);
							} else if (node.getText().equals("state")) {
								stateIdentNameToVarNameMap.put(identName, varName);
							} else if (node.getText().equals("action")) {
								actionIdentNameToVarNameMap.put(identName, varName);
							}
						} else {
							var = new CommonTree(new CommonToken(DuraParser.VARIABLE, "VARIABLE"));
							var.addChild(new CommonTree(new CommonToken(DuraParser.ID, varName)));
						}
						
						index = node.getChildIndex();
						node.getParent().replaceChildren(index, index, var);
					}
					continue;
					
				case DuraParser.VARIABLE:
					var = node;
					varName = var.getChild(0).getText();
					
					if (varNameToTypeNodesMap.get(varName) == null) {
						
						varNameToExpressionMap.put(varName, copyTree(var));
						
						schemaTerm = actionSchemataMap.get(actionName);
						typeNodes = getTypeNodes(schemaTerm, node);
						
						varNameToTypeNodesMap.put(varName, typeNodes);
					}
					
					continue;
				}
			} else if (node.getParent().getType() == DuraParser.ACTION_EXECUTION && node.getType() == DuraParser.TERM) {
				actionName = node.getText();
			}
			
			if (node.getChildCount() > 0) {
				for (int i = node.getChildCount() - 1; i >= 0; i--) {
					CommonTree child = (CommonTree) node.getChild(i);	
					
					if (child.getType() != DuraParser.EXECUTION_CONSTRAINTS &&
						child.getType() != DuraParser.SUCCEEDS_ON &&
						child.getType() != DuraParser.FAILS_ON) {
						
						nodeStack.push(child);
					}
				}
			}
		}
	}
}
