package de.lmu.ifi.pms.cep.event_mill.sample;

import java.io.IOException;
import java.util.ArrayList;

import de.lmu.ifi.pms.cep.event_mill.EventMill;
import de.lmu.ifi.pms.cep.event_mill.configuration.Configuration;
import de.lmu.ifi.pms.cep.event_mill.configuration.IOConfiguration;
import de.lmu.ifi.pms.cep.event_mill.execution.engine.Engine;
import de.lmu.ifi.pms.cep.event_mill.execution.program.ExecutableProgram;
import de.lmu.ifi.pms.cep.event_mill.execution.scheduler.SchedulerFactory;
import de.lmu.ifi.pms.cep.event_mill.io.MessageSink;
import de.lmu.ifi.pms.cep.event_mill.io.MessageSource;
import de.lmu.ifi.pms.cep.event_mill.io.PATTERNS;
import de.lmu.ifi.pms.cep.event_mill.io.bus.MessageBus;
import de.lmu.ifi.pms.cep.event_mill.io.connect.BusToStreamConnector;
import de.lmu.ifi.pms.cep.event_mill.io.connect.StreamToBusConnector;
import de.lmu.ifi.pms.cep.event_mill.io.format.SemiStructuredMessageFormatProvider;
import de.lmu.ifi.pms.cep.event_mill.io.format.XmlMessageFormatProvider;
import de.lmu.ifi.pms.cep.event_mill.io.stream.StreamIO;
import de.lmu.ifi.pms.cep.tsa.DTSA_Program;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;

public class Main {
	
	public static void main(final String[] args) throws Exception{
		
		String sourceFolder = "examples/";
		String module = "example.hello-world";
		String dtsaFile = "examples/hello-world.xml";
		String instance = "test";
		
		final String instanceFullName = TSA_SyntaxPatterns.programInstFullName(module, instance);
		
		String[] schemas = new String[]{
								"sample_meta", 
								"sample_input", 
								"sample_working", 
								"sample_output", 
								"sample_log"
							   };
		
		
		
		EventMill eventMill = new EventMill(Configuration.read());
		
		//clear event mill
		eventMill.clear();
		
		DTSA_Program dtsa_program = eventMill.compile(module, sourceFolder);
		
		assert dtsa_program.full_name.equals(module);
		
		eventMill.save(dtsa_program, dtsaFile);		
		dtsa_program = eventMill.load(dtsaFile);
		
		eventMill.addSpec(dtsa_program);
		
		//list available program specifications
		System.out.println("Available specifications:");
		for(String spec: eventMill.availableProgramSpecifications()){
			System.out.println(spec);
		}
		System.out.println();
		
		eventMill.instanciate(module, instance, schemas);
		
		//list available program instances
		System.out.println("Available program instances:");
		for(String spec: eventMill.availablePrograms()){
			System.out.println(spec);
		}
		System.out.println();
		
		//Obtain program instance
		ExecutableProgram<?, ?, ?> executableProgram = eventMill.getProg(module, instance);
		
		//Create engine for program instance
		Engine engine = new Engine(executableProgram);
		
		
		//Connect Engine to message bus
		
		IOConfiguration ioConfiguration;
		try {
			ioConfiguration = IOConfiguration.read();
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	
		String engineIdentifier = "de.lmu.ifi.pms.cep.event-mill.prototype";
		
		String programInstanceIdentifier = PATTERNS.programIdentifier(engineIdentifier, instanceFullName);
		
		StreamIO streamIO = ioConfiguration.stream(engineIdentifier, programInstanceIdentifier);
		
		String programInstanceBusEndpoint = programInstanceIdentifier;
		final Class<String> messageType = String.class;
		
		final String inputBusIdentifier = PATTERNS.inputBusIdentifier(engineIdentifier, instanceFullName);
		String inputMessageFormatIdentifier = SemiStructuredMessageFormatProvider.defaultIdentifier;
		MessageBus<String> inputBus = ioConfiguration.bus(inputBusIdentifier, messageType, inputMessageFormatIdentifier, programInstanceBusEndpoint);
		
		final String outputBusIdentifier = PATTERNS.outputBusIdentifier(engineIdentifier, instanceFullName);
		String outputMessageFormatIdentifier = XmlMessageFormatProvider.defaultIdentifier;
		MessageBus<String> outputBus = ioConfiguration.bus(outputBusIdentifier, messageType, outputMessageFormatIdentifier, programInstanceBusEndpoint);
		
		final BusToStreamConnector<?> inputConnector = BusToStreamConnector.create(inputBus, streamIO);
		final StreamToBusConnector<?> outputConnector = StreamToBusConnector.create(outputBus, streamIO);
		
		Thread ioThread = new Thread("IO-Thread"){
			
			@Override
			public void run() {				
				while(!this.isInterrupted()){
					inputConnector.nextBlock();
					outputConnector.nextBlock();
					try {
						Thread.sleep(1000);
					} catch (InterruptedException e) {
						break;
					}
				}
				
			}
			
		};
		
		//Thread for providing input
		Thread simInputThread = new Thread("Emulation"){
			
			@Override
			public void run() {
				
				String emulationBusEndPoint = "de.lmu.ifi.pms.cep.event_mill.sample.emulation";
				String emulationMessageFormatIdentifier = SemiStructuredMessageFormatProvider.defaultIdentifier;
				MessageBus<String> emulationOutputBus;

				try {
					
					emulationOutputBus = IOConfiguration.read().bus(inputBusIdentifier, messageType, emulationMessageFormatIdentifier, emulationBusEndPoint);
				} catch (IOException e) {
					throw new RuntimeException(e);
				}
				
				MessageSink<String> messageSink = emulationOutputBus.sink();
				
				ArrayList<String> messages = new ArrayList<String>();
				
				int round = 1;
				while(!this.isInterrupted()){
					messages.add("example.hello-world.see-person{ person{ \"boy_" + round + "\" } }");
					messages.add("example.hello-world.see-person{ person{ \"girl_" + round + "\" } }");
					
					messageSink.writeBlock(messages);
					
					messages.clear();
					
					try {
						Thread.sleep(1000);
					} catch (InterruptedException e) {
						break;
					}
					++round;
				}
			}
			
		};
		
		//Thread for reading output 
		Thread uiThread = new Thread("UI"){
			
			@Override
			public void run() {
				
				String uiBusEndPoint = "de.lmu.ifi.pms.cep.event_mill.sample.ui";
				String uiMessageFormatIdentifier = SemiStructuredMessageFormatProvider.defaultIdentifier;
				MessageBus<String> uiInputBus;
				
				try {
					uiInputBus = IOConfiguration.read().bus(outputBusIdentifier, messageType, uiMessageFormatIdentifier, uiBusEndPoint);
				} catch (IOException e) {
					throw new RuntimeException(e);
				}
				
				MessageSource<String> messageSource = uiInputBus.source();
				
				while(!this.isInterrupted()){					
					for(String message: messageSource.readBlock()){
						System.out.println(message);
					}
					
					try {
						Thread.sleep(1000);
					} catch (InterruptedException e) {
						break;
					}
				}
			}
			
		};
		
		//initialize engine
		
		SchedulerFactory scheduler = eventMill.configuration().scheduler;
		//retrieving the scheduler from the configuration of the Event-Mill instance
		
		engine.init(scheduler);
		//initializing the engine instance for the given program with the assigned scheduler
		
		//start program
		
		engine.start();
		
		//start IO
		simInputThread.start();
		ioThread.start();
		uiThread.start();
		
		//wait for user input
		System.out.println("Press [Enter] to stop execution.");
		System.in.read();
		
		//stop IO
		simInputThread.interrupt();
		simInputThread.join();
		ioThread.interrupt();
		ioThread.join();
		uiThread.interrupt();
		uiThread.join();
		
		//stop program
		engine.stop();
		
		//release resources
		engine.close();
		
		//remove program instance
		eventMill.removeProg(module, instance);
		
		//remove program specification
		eventMill.removeSpec(module);
		
		eventMill.close();		
	}

}
