package de.lmu.ifi.pms.cep.event_mill.sql;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import de.lmu.ifi.pms.cep.COMMON_CONSTANTS;
import de.lmu.ifi.pms.cep.event_mill.execution.program.ExecutableProgram;
import de.lmu.ifi.pms.cep.event_mill.execution.program.ExecutableProgramState;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.DTSA_QueryType;
import de.lmu.ifi.pms.cep.tsa.time.Clock;
import de.lmu.ifi.pms.cep.tsa.time.SystemClock;

import de.lmu.ifi.pms.java.util.collection.array.Arrays;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArraySet;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;
import de.lmu.ifi.pms.java.util.map.fixed.FixedEnumMap;


public class SQL_ExecutableProgram extends ExecutableProgram<SQL_StreamBufferStore, SQL_StreamBuffer, SQL_ExecutableQuery>
{
	final int id;
	
	public final SQL_Translation sql_translation;
	
	final String meta_schema;
	final String input_schema;
	final String working_schema;
	final String output_schema;
	final String log_schema;
	final FixedEnumMap<DTSA_BufferType, String> bufferSchemas;
	
	long[] minTimesEraser;
	
	FixedArrayNavigableMap<String, Clock> clocks;
	
	FixedEnumMap<DTSA_BufferType,FixedArrayNavigableMap<String, SQL_StreamBuffer>> streamBuffers;
	FixedArraySet<SQL_ExecutableQuery>[] queries;
	
	int bufferMaxNumberOfMinTimes = 0;
	int queryMaxNumberOfOutputMinTimes = 0;
	int queryMaxNumberOfInputMinTimes = 0;
	int queryMaxNumberOfReferencedClocks = 0;
	
	protected boolean usePreparedStatements;
	
	protected final String setStateStatement;
	protected PreparedStatement sqlSetStateStatement;
	protected final String getStateStatement;
	protected PreparedStatement sqlGetStateStatement;
	protected final String aquireLockStatement;
	protected PreparedStatement sqlAquireLockStatement;
	protected final String releaseLockStatement;
	protected PreparedStatement sqlReleaseLockStatement;
	
	@SuppressWarnings("unchecked")
	SQL_ExecutableProgram(final int id, final String namespace, final String simple_name, final String instance, final String meta_schema, final String input_schema, final String working_schema, final String output_schema, final String log_schema, final SQL_StreamBufferStore bufferStore)
	{
		super(namespace, simple_name, instance, bufferStore);
		this.id = id;
		this.sql_translation = bufferStore.sql_translation;
		
		this.meta_schema = meta_schema;
		this.input_schema = input_schema;
		this.working_schema = working_schema;
		this.output_schema = output_schema;
		this.log_schema = log_schema;
		this.bufferSchemas = new FixedEnumMap<DTSA_BufferType, String>(DTSA_BufferType.class, new String[]{this.input_schema, this.working_schema, this.output_schema, this.log_schema});
		
		this.minTimesEraser = null;
		
		this.usePreparedStatements = bufferStore.usePreparedStatements;
		
		this.clocks = new FixedArrayNavigableMap<String, Clock>(new String[]{COMMON_CONSTANTS.SYSTEM_CLOCK_NAME, COMMON_CONSTANTS.TUPLE_ID_CLOCK_NAME}, new Clock[]{new SystemClock(), new TupleIdSequenceClock(this.bufferStore, this.meta_schema, SQL_Translation.tupleIdSeqenceName, this.usePreparedStatements)});
		
		FixedArrayNavigableMap<String, SQL_StreamBuffer>[] streamBuffers = new FixedArrayNavigableMap[DTSA_BufferType.count];
		for(int ordinal = 0; ordinal < DTSA_BufferType.count; ++ordinal){
			streamBuffers[ordinal] = new FixedArrayNavigableMap<String, SQL_StreamBuffer>();
		}
		this.streamBuffers = new FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String,SQL_StreamBuffer>>(DTSA_BufferType.class, streamBuffers);
		
		this.queries = (FixedArraySet<SQL_ExecutableQuery>[]) new FixedArraySet<?>[DTSA_QueryType.values().length];
		java.util.Arrays.fill(this.queries, new FixedArraySet<SQL_ExecutableQuery>());
		
		this.setStateStatement = this.sql_translation.sqlSetProgramInstState(this.bufferStore.meta_schema_name, this.id);
		this.getStateStatement = this.sql_translation.sqlGetProgramInstState(this.bufferStore.meta_schema_name, this.id);
		this.aquireLockStatement = this.sql_translation.sqlLockProgramInst(this.bufferStore.meta_schema_name, this.id);
		this.releaseLockStatement = this.sql_translation.sqlUnlockProgramInst(this.bufferStore.meta_schema_name, this.id);
		this.initStatements();
		
	}
	
	@Override
	protected synchronized boolean aquireLock() {
		while(true){
			try {
				int updateCount = this.sqlAquireLockStatement.executeUpdate();
				assert updateCount == 0 | updateCount == 1;
				boolean locked = updateCount > 0;
				this.debug("Lock statement returned with: " + updateCount);
				return locked;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					try {
						this.bufferStore.reconnect(this.sqlSetStateStatement.getConnection());
					} catch (SQLException e1) {
						this.error("", e);
						throw new RuntimeException(e);
					}
					this.initStatements();
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Aquire lock statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of aquire lock statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
					//TODO only init the affected statement?
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
					//TODO better exception
				}
			}
		}
	}
	
	@Override
	protected synchronized void releaseLock() {
		while(true){
			try {
				int updateCount = this.sqlReleaseLockStatement.executeUpdate();
				assert updateCount == 0 | updateCount == 1;
				boolean unlocked = updateCount > 0;
				this.debug("Unlock statement returned with: " + updateCount);
				break;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					try {
						this.bufferStore.reconnect(this.sqlSetStateStatement.getConnection());
					} catch (SQLException e1) {
						this.error("", e);
						throw new RuntimeException(e);
					}
					this.initStatements();
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Aquire lock statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of aquire lock statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
					//TODO only init the affected statement?
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
					//TODO better exception
				}
			}
		}
	}
	
	@Override
	protected ExecutableProgramState readState() {
		return this.sqlGetState();
	}
	
	protected ExecutableProgramState sqlGetState(){
		while(true){
			try {
				ResultSet state = this.sqlGetStateStatement.executeQuery();
				boolean existsProgram = state.next();
				assert existsProgram;
				return this.sql_translation.sqlProgramInstReadState(state);
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Get state statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of get state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
					//TODO only init the affected statement?
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
					//TODO better exception
				}
			}
		}
	}
	
	protected void sqlSetState(final ExecutableProgramState state){
		while(true){
			try {
				this.sql_translation.sqlSetSetProgramInstState(this.sqlSetStateStatement, state);
				int updateCount = this.sqlSetStateStatement.executeUpdate();
				assert updateCount == 1;
				break;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					try {
						this.bufferStore.reconnect(this.sqlSetStateStatement.getConnection());
					} catch (SQLException e1) {
						this.error("", e);
						throw new RuntimeException(e);
					}
					this.initStatements();
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Set state statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
					//TODO only init the affected statement?
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
					//TODO better exception
				}
			}
		}
	}
	
	protected void initStatements(){
		int tryCount = 2;
		
		while(true){
			Connection connection = this.bufferStore.connection();
			try
			{
				this.sqlGetStateStatement = this.sql_translation.prepareStatement(connection, this.getStateStatement, this.usePreparedStatements);
				this.sqlSetStateStatement = this.sql_translation.prepareStatement(connection, this.setStateStatement, this.usePreparedStatements);
				this.sqlAquireLockStatement = this.sql_translation.prepareStatement(connection, this.aquireLockStatement, this.usePreparedStatements);
				this.sqlReleaseLockStatement = this.sql_translation.prepareStatement(connection, this.releaseLockStatement, this.usePreparedStatements);
				break;
			} catch (SQLException e)
			{
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.bufferStore.reconnect(connection);
				}
				else if(SQL_EXCEPTIONS.isMonetDBBug(e) & tryCount > 0){
					this.warn("MonetDB bug. Trying to reconnect.", e);
					this.bufferStore.reconnect(connection);
					--tryCount;
				}
				else{
					this.error("Error on creating statement!", e);
					throw new RuntimeException(e);
					//TODO better exception
				}
			}
		}
	}
	
	@Override
	public synchronized void init() {
		super.init();
		this.sqlSetState(this.state);
	}
	
	@Override
	public synchronized void lock() {
		super.lock();
		assert this.state != null;
		this.minTimesEraser = new long[Math.max(this.bufferMaxNumberOfMinTimes, Math.max(this.queryMaxNumberOfOutputMinTimes, this.queryMaxNumberOfInputMinTimes))];
		Arrays.fill(this.minTimesEraser, Long.MAX_VALUE);
	}
	
	@Override
	public synchronized void unlock() {
		super.unlock();
		this.minTimesEraser = null;
	}
	
	@Override
	public synchronized void run() {		
		super.run();
		this.sqlSetState(this.state);
	}
	
	@Override
	public synchronized void stop() {
		super.stop();
		this.sqlSetState(this.state);
	}
	
	
	synchronized void erase(final long[] minTimes){
		assert this.isLocked();
		assert this.minTimesEraser != null;
		assert minTimes.length <= this.minTimesEraser.length;
		System.arraycopy(this.minTimesEraser, 0, minTimes, 0, minTimes.length);
	}
	

	@Override
	public FixedEnumMap<DTSA_BufferType,FixedArrayNavigableMap<String, SQL_StreamBuffer>> StreamBuffers()
	{
		return this.streamBuffers;
	}
	
	synchronized void register(final SQL_StreamBuffer buffer){
		if(!this.isLocked()){			
			String bufferName = buffer.name();
			DTSA_BufferType bufferType = buffer.type;
			FixedArrayNavigableMap<String, SQL_StreamBuffer> streamBuffersOfSameType = this.streamBuffers.get(bufferType);
			int index = streamBuffersOfSameType.keys().ceilingIndex(bufferName);
			if(index < streamBuffersOfSameType.size() && bufferName.equals(streamBuffersOfSameType.keys().get(index))){
				Error e = new Error("Invalid call to register. Already registered buffer with type \"" + buffer.type + "\" and name \"" + bufferName + "\".");
				
				this.error("", e); //TODO better log message
			}
			else{
				String[] bufferNames = new String[streamBuffersOfSameType.size() + 1];
				SQL_StreamBuffer[] buffers = new SQL_StreamBuffer[bufferNames.length];
				
				streamBuffersOfSameType.keys().copyTo(0, bufferNames, 0, index);
				streamBuffersOfSameType.values().copyTo(0, buffers, 0, index);
				bufferNames[index] = bufferName;
				buffers[index] = buffer;
				streamBuffersOfSameType.keys().copyTo(index, bufferNames, index + 1, streamBuffersOfSameType.size() - index);
				streamBuffersOfSameType.values().copyTo(index, buffers, index + 1, streamBuffersOfSameType.size() - index);

				@SuppressWarnings("unchecked")
				FixedArrayNavigableMap<String, SQL_StreamBuffer>[] streamBuffers = new FixedArrayNavigableMap[DTSA_BufferType.count];
				for(int ordinal = 0; ordinal < DTSA_BufferType.count; ++ordinal){
					if(ordinal == bufferType.ordinal()){
						streamBuffers[ordinal] = new FixedArrayNavigableMap<String, SQL_StreamBuffer>(new FixedArrayNavigableSet<String>(false, false, bufferNames), new FixedArray<SQL_StreamBuffer>(false, buffers));
					}
					else{
						streamBuffers[ordinal] = this.streamBuffers.values().get(ordinal);
					}
				}
				
				this.streamBuffers = new FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String,SQL_StreamBuffer>>(DTSA_BufferType.class, streamBuffers);
				
				if(buffer.numberOfMinTimes > this.bufferMaxNumberOfMinTimes){
					this.bufferMaxNumberOfMinTimes = buffer.numberOfMinTimes;
				}
				
				this.clocks = FixedArrayNavigableMap.merge(this.clocks, new FixedArrayNavigableMap<String, Clock>(COMMON_CONSTANTS.BLOCK_COUNTER_CLOCK_NAME(buffer.type, buffer.full_name),new BufferBlockCounterClock(this.bufferStore, this.meta_schema, buffer.type, buffer.full_name, this.usePreparedStatements)));
				
				this.debug("Registered Buffer {} {}", buffer.id, buffer.name()); //TODO better log message
			}
		}
		else{
			throw new IllegalStateException();
		}
	}

	@Override
	public FixedEnumMap<DTSA_QueryType, FixedArraySet<SQL_ExecutableQuery>> Queries()
	{
		return new FixedEnumMap<DTSA_QueryType, FixedArraySet<SQL_ExecutableQuery>>(DTSA_QueryType.class, java.util.Arrays.copyOf(this.queries, this.queries.length));
	}
	
	synchronized void register(final SQL_ExecutableQuery query){
		if(!this.isLocked()){
			SQL_ExecutableQuery[] queryTypeQueries = new SQL_ExecutableQuery[this.queries[query.type.ordinal()].size() + 1];
			for(int index = 0; index < this.queries[query.type.ordinal()].size(); ++index){
				if(query == this.queries[query.type.ordinal()].get(index)){//TODO better test?
					Error e = new Error("Invalid call to register.");
					
					this.error("", e); //TODO better log message
				}
				queryTypeQueries[index] = this.queries[query.type.ordinal()].get(index);
			}
			queryTypeQueries[queryTypeQueries.length - 1] = query;
			this.queries[query.type.ordinal()] = new FixedArraySet<SQL_ExecutableQuery>(false, false, queryTypeQueries);
			
			if(query.numberOfOutputMinTimes > this.queryMaxNumberOfOutputMinTimes){
				this.queryMaxNumberOfOutputMinTimes = query.numberOfOutputMinTimes;
			}
			if(query.numberOfInputMinTimes > this.queryMaxNumberOfInputMinTimes){
				this.queryMaxNumberOfInputMinTimes = query.numberOfInputMinTimes;
			}		
			if(query.numberOfReferencedClocks > this.queryMaxNumberOfReferencedClocks){
				this.queryMaxNumberOfReferencedClocks = query.numberOfReferencedClocks;
			}
			
			this.debug("Registered Query {}", query.id); //TODO better log message
		}
		else{
			throw new IllegalStateException();
		}
	}
	
	
//	protected static HashMap<String, HashMap<String,WeakReference<SQL_StreamBufferStore>>> stores = new HashMap<String, HashMap<String,WeakReference<SQL_StreamBufferStore>>>();
//	
//	protected static synchronized void remove(final String url, final String catalog, final String meta_schema_name){
//		stores.get(url).remove(encode(catalog, meta_schema_name));
//		if(stores.get(url).isEmpty()){
//			stores.remove(url);;
//		}
//	}
//	
//	protected static synchronized void put(final SQL_StreamBufferStore store){
//		if(!stores.containsKey(store.url)){
//			stores.put(store.url, new HashMap<String, WeakReference<SQL_StreamBufferStore>>());
//		}
//		stores.get(store.url).put(encode(store.catalog, store.meta_schema_name), new WeakReference<SQL_StreamBufferStore>(store));
//	}

}
