package de.lmu.ifi.pms.cep.event_mill.sql;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Arrays;
import java.util.concurrent.locks.ReentrantLock;
import java.util.regex.Matcher;

import de.lmu.ifi.pms.cep.event_mill.execution.program.ExecutableQuery;
import de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper;
import de.lmu.ifi.pms.cep.event_mill.sql.optimization.dtsa.DTSA_RecoveryEnhacer;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.DTSA_QueryType;
import de.lmu.ifi.pms.cep.tsa.time.Clock;
import de.lmu.ifi.pms.cep.tsa.time.functions.TimeFunction;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;
import de.lmu.ifi.pms.java.util.map.fixed.FixedEnumMap;


public class SQL_ExecutableQuery extends ExecutableQuery<SQL_ExecutableProgram, SQL_StreamBuffer>
{
	
	public final int id;
	public final SQL_Translation sql_translation;
	
	public final String id_attribute;
	
	long lastRound;
	public final String round_attribute;
	//first value = id, second value = round
	protected final long[] queryInfo = new long[2];
	
	protected final int numberOfReferencedClocks;
	protected final FixedArrayNavigableMap<String, Clock> referencedClocks;
	protected final int numberOfOutputMinTimes;
	protected long[] lastOutputMinTimes;
	protected long[] nextOutputMinTimes;
	protected long[] lastOutputMinWriteTimes;
	protected long[] nextOutputMinWriteTimes;

	protected final int numberOfInputMinTimes;
	protected long[] lastInputMinTimes;
	protected long[] nextInputMinTimes;
	protected final long[] clockTimes;
	protected final long[] clockMinTimes;
	protected final long[] clockMinTimeIncreases;
	protected final TimeFunction minTimesFunction;
	protected final TimeFunction minTimeClockIncreasesFunction;
	
	protected final ReentrantLock executionLock = new ReentrantLock();
	protected boolean chrashed = false;
	protected final Object minWriteTimesLock = new Object();
	
	protected boolean usePreparedStatements;
	
	protected final String eval_statement;
	protected boolean areSetEvalParameters = false;
	protected PreparedStatement sql_eval_statement;
	
	protected static enum ParameterSource{
		InputLastMinTimes,
		InputNowMinTimes,
		OutputLastMinTimes,
		OutputNowMinTimes,
		ClockNowTimes,
		QueryInfo
	}
	
	protected final int numberOfParameters;
	protected final ParameterSource[] parameterMappingSource;
	protected final int[] parameterMappingLocalIndex;
	
	protected final String updateStatement;
	protected boolean areSetMetaParameters = false;
	protected PreparedStatement MetaEntryUpdate;

	protected final long[] lastTupleCounts;
	protected final FixedArrayNavigableMap<String, SQL_StreamBuffer> inputs;
	
	protected SQL_StreamBuffer output;
	
	protected final boolean isMonotonic;
	
	protected Statement loggingStatement;
	
	public SQL_ExecutableQuery(final SQL_ExecutableProgram program, final int id, final DTSA_QueryType type, final String prog_package, final boolean isMonotonic, final String sql_execution_statement, final FixedArrayNavigableSet<String> referencedClocks, final TimeFunction minTimesFunction, final TimeFunction minTimeClockIncreasesFunction, final DTSA_BufferType input_type, final FixedArrayNavigableSet<String> inputs, final String output, final long lastRound, final long[] inputMinTimes){
		this(program, id, type, prog_package, isMonotonic, sql_execution_statement, referencedClocks, minTimesFunction, minTimeClockIncreasesFunction, input_type, inputs, output, lastRound, inputMinTimes, program.usePreparedStatements);
	}

	public SQL_ExecutableQuery(final SQL_ExecutableProgram program, final int id, final DTSA_QueryType type, final String prog_package, final boolean isMonotonic, final String sql_execution_statement, final FixedArrayNavigableSet<String> referencedClocks, final TimeFunction minTimesFunction, final TimeFunction minTimeClockIncreasesFunction, final DTSA_BufferType input_type, final FixedArrayNavigableSet<String> inputs, final String output, final long lastRound, final long[] inputMinTimes, final boolean usePreparedStatements)
	{
		super(type, prog_package, program);
		FixedEnumMap<DTSA_BufferType,FixedArrayNavigableMap<String, SQL_StreamBuffer>> availableStreamBuffers = program.StreamBuffers();
		FixedArrayNavigableMap<String, Clock> availableClocks = program.clocks;
		
		
		this.id = id;
		this.sql_translation = program.sql_translation;
		
		//FIXME retrieve name of sql attribute generically
		this.id_attribute = SQL_Translation.SQL_BasicAttributeName(DTSA_RecoveryEnhacer.tsa_recovery_query_id);
		
		this.lastRound = lastRound;
		//FIXME retrieve name of sql attribute generically
		this.round_attribute = SQL_Translation.SQL_BasicAttributeName(DTSA_RecoveryEnhacer.tsa_recovery_round);
		
		DTSA_BufferType outputBufferType = SQL_Translation.outputBufferType(this.type);
		
		int outputStreamBufferIndex = availableStreamBuffers.get(outputBufferType).keys().indexOf(output);
		if(outputStreamBufferIndex >= 0){
			this.output = availableStreamBuffers.get(outputBufferType).values().get(outputStreamBufferIndex);
		}
		else{
			throw new RuntimeException("Query " + this.id + "requires buffer \"" + output + "\" of type \"" + outputBufferType.name() + "\"that is not available in the program.");
		}
		
		this.output.RegisterInputQuery(this);
		
		this.numberOfOutputMinTimes = this.output.numberOfMinTimes;
		
		this.lastOutputMinTimes = new long[this.numberOfOutputMinTimes];
		this.nextOutputMinTimes = new long[this.numberOfOutputMinTimes];
		this.lastOutputMinWriteTimes = new long[this.numberOfOutputMinTimes];
		this.nextOutputMinWriteTimes = new long[this.numberOfOutputMinTimes];
		
		if(!SQL_Translation.isPossibleInputBuffersType(type, input_type)){
			throw new RuntimeException("The type of query " + this.id + " and its specified input type are incompatible.");
		}
		
		SQL_StreamBuffer[] inputBuffers = new SQL_StreamBuffer[inputs.size()];
		for(int index = 0; index < inputBuffers.length; ++index){
			String input = inputs.get(index);
			int streamBufferIndex = availableStreamBuffers.get(input_type).keys().indexOf(input);
			if(streamBufferIndex >= 0){
				inputBuffers[index] = availableStreamBuffers.get(input_type).values().get(streamBufferIndex);
			}
			else{
				throw new RuntimeException("Query " + this.id + "requires buffer \"" + input + "\", but \"" + input + "\" is not available in the program.");
			}
		}		
		this.inputs = new FixedArrayNavigableMap<String, SQL_StreamBuffer>(inputs, new FixedArray<SQL_StreamBuffer>(false, inputBuffers));
		for(SQL_StreamBuffer input: this.inputs.values()){
			input.RegisterOutputQuery(this);
		}
		this.lastTupleCounts = new long[inputs.size()];
		Arrays.fill(this.lastTupleCounts, Long.MIN_VALUE);//Will lead to at most one unnecessary query execution
		
		int numberOfInputMinTimes = 0;
		int[] numberOfPreceedingMinTimes = new int[inputs.size()];
		for(int index = 0; index < numberOfPreceedingMinTimes.length; ++index){
			SQL_StreamBuffer inputBuffer = this.inputs.values().get(index);
			numberOfPreceedingMinTimes[index] = numberOfInputMinTimes;
			
			numberOfInputMinTimes += inputBuffer.numberOfMinTimes;
		}

		this.numberOfInputMinTimes = numberOfInputMinTimes;
		if(!(this.numberOfInputMinTimes == inputMinTimes.length)){
			throw new IllegalArgumentException();
			//TODO better exception
		}
		
		this.lastInputMinTimes = new long[this.numberOfInputMinTimes];
		this.nextInputMinTimes = new long[this.numberOfInputMinTimes];
		
		System.arraycopy(inputMinTimes, 0, this.lastInputMinTimes, 0, this.numberOfInputMinTimes);
		
		
		this.numberOfReferencedClocks = referencedClocks.size();
		Clock[] clocks = new Clock[this.numberOfReferencedClocks];
		for(int clock = 0; clock < this.numberOfReferencedClocks; ++ clock){
			if(availableClocks.has(referencedClocks.get(clock))){
				clocks[clock] = availableClocks.get(referencedClocks.get(clock));
			}
			else{
				throw new IllegalArgumentException("Query " + this.id + "references unknown clock: " + referencedClocks.get(clock));
				//TODO better exception
			}
		}
		
		this.referencedClocks = new FixedArrayNavigableMap<String, Clock>(referencedClocks, new FixedArray<Clock>(false, clocks));
		
		this.clockTimes = new long[this.numberOfReferencedClocks];
		this.clockMinTimes = new long[this.numberOfReferencedClocks];
		this.clockMinTimeIncreases = new long[this.numberOfOutputMinTimes];
		
		this.minTimesFunction = minTimesFunction;
		assert this.minTimesFunction.argumentArity() == this.numberOfInputMinTimes;
		assert this.minTimesFunction.resultArity() == this.numberOfOutputMinTimes: "Query " + this.id + ": The query has to compute exactly as many outputMinTimes as required by its output type.";
		this.minTimeClockIncreasesFunction = minTimeClockIncreasesFunction;
		assert this.minTimeClockIncreasesFunction.argumentArity() == this.numberOfReferencedClocks;
		assert this.minTimeClockIncreasesFunction.resultArity() == this.numberOfOutputMinTimes;

		this.minTimesFunction.apply(this.lastInputMinTimes, 0, this.lastOutputMinTimes, 0);
		this.updateOutputMinWriteTimes();
		
		this.queryInfo[0] = this.id;
		this.queryInfo[1] = this.lastRound;
		
		String[] parameters = this.sql_translation.parametersOfInstanciatedSQLStatement(sql_execution_statement);
		//this.parameter_splited_eval_statement = sql_translation.parameterSplitedInstanciatedSQLStatement(sql_execution_statement);

		this.numberOfParameters = parameters.length;
		this.parameterMappingSource = new ParameterSource[this.numberOfParameters];
		this.parameterMappingLocalIndex = new int[this.numberOfParameters];
		
		for(int parameterIndex = 0; parameterIndex < this.numberOfParameters; ++parameterIndex){
			String parameter = parameters[parameterIndex];
			assert parameter.matches(SQL_Translation.parameter_qualified_regEx);
			
			Matcher parameterMatcher = SQL_Translation.parameter_qualified_Pattern.matcher(parameter);
			parameterMatcher.matches();
			
			int[] groupIndexOf = new int[]{1,6,11,16,21,23};
			if(parameterMatcher.group(groupIndexOf[ParameterSource.InputLastMinTimes.ordinal()]) != null){
				int groupIndex =  groupIndexOf[ParameterSource.InputLastMinTimes.ordinal()];
				assert parameterMatcher.group(groupIndex).equals("input.last");
				++groupIndex;
				
				int buffer_index = inputs.indexOf(parameterMatcher.group(groupIndex));
				assert buffer_index >= 0;
				groupIndex += 3;
				int minTime_index = this.inputs.values().get(buffer_index).min_times.indexOf(parameterMatcher.group(groupIndex));
				if(minTime_index < 0){
					throw new IllegalArgumentException("Unknown min-time \"" + parameterMatcher.group(groupIndex)+ "\"");
				}
				
				this.parameterMappingSource[parameterIndex] = ParameterSource.InputLastMinTimes;
				this.parameterMappingLocalIndex[parameterIndex] = numberOfPreceedingMinTimes[buffer_index] + minTime_index;
				assert this.parameterMappingLocalIndex[parameterIndex] >= 0; 
				++groupIndex;
			}
			else if(parameterMatcher.group(groupIndexOf[ParameterSource.InputNowMinTimes.ordinal()]) != null){
				int groupIndex =  groupIndexOf[ParameterSource.InputNowMinTimes.ordinal()];
				assert parameterMatcher.group(groupIndex).equals("input.now");
				++groupIndex;
				
				int buffer_index = inputs.indexOf(parameterMatcher.group(groupIndex));
				assert buffer_index >= 0;
				groupIndex += 3;
				int minTime_index = this.inputs.values().get(buffer_index).min_times.indexOf(parameterMatcher.group(groupIndex));
				if(minTime_index < 0){
					throw new IllegalArgumentException("Unknown min-time \"" + parameterMatcher.group(groupIndex)+ "\"");
				}
				
				this.parameterMappingSource[parameterIndex] = ParameterSource.InputNowMinTimes;
				this.parameterMappingLocalIndex[parameterIndex] = numberOfPreceedingMinTimes[buffer_index] + minTime_index;
				assert this.parameterMappingLocalIndex[parameterIndex] >= 0;
				++groupIndex;
			}
			else if(parameterMatcher.group(groupIndexOf[ParameterSource.OutputLastMinTimes.ordinal()]) != null){
				int groupIndex =  groupIndexOf[ParameterSource.OutputLastMinTimes.ordinal()];
				assert parameterMatcher.group(groupIndex).equals("output.last");
				++groupIndex;
				
				assert parameterMatcher.group(groupIndex).equals(this.output.name());
				groupIndex += 3;
				int minTime_index = this.output.min_times.indexOf(parameterMatcher.group(groupIndex));
				if(minTime_index < 0){
					throw new IllegalArgumentException("Unknown min-time \"" + parameterMatcher.group(groupIndex)+ "\"");
				}
				
				this.parameterMappingSource[parameterIndex] = ParameterSource.OutputLastMinTimes;
				this.parameterMappingLocalIndex[parameterIndex] = minTime_index;
				assert this.parameterMappingLocalIndex[parameterIndex] >= 0;
				++groupIndex;
			}
			else if(parameterMatcher.group(groupIndexOf[ParameterSource.OutputNowMinTimes.ordinal()]) != null){
				int groupIndex =  groupIndexOf[ParameterSource.OutputNowMinTimes.ordinal()];
				assert parameterMatcher.group(groupIndex).equals("output.now");
				++groupIndex;
				
				assert parameterMatcher.group(groupIndex).equals(this.output.name());
				groupIndex += 3;
				int minTime_index = this.output.min_times.indexOf(parameterMatcher.group(groupIndex));
				if(minTime_index < 0){
					throw new IllegalArgumentException("Unknown min-time \"" + parameterMatcher.group(groupIndex)+ "\"");
				}
				
				this.parameterMappingSource[parameterIndex] = ParameterSource.OutputNowMinTimes;
				this.parameterMappingLocalIndex[parameterIndex] = minTime_index;
				assert this.parameterMappingLocalIndex[parameterIndex] >= 0;
				++groupIndex;
			}
			else if(parameterMatcher.group(groupIndexOf[ParameterSource.ClockNowTimes.ordinal()]) != null){
				int groupIndex =  groupIndexOf[ParameterSource.ClockNowTimes.ordinal()];
				assert parameterMatcher.group(groupIndex).equals("clock.now");
				++groupIndex;
				
				int clock_index = this.referencedClocks.keys().indexOf(parameterMatcher.group(groupIndex));
				if(clock_index < 0){
					throw new IllegalArgumentException("Unknown clock \"" + parameterMatcher.group(groupIndex)+ "\"");
				}
				
				this.parameterMappingSource[parameterIndex] = ParameterSource.ClockNowTimes;
				this.parameterMappingLocalIndex[parameterIndex] = clock_index;
				++groupIndex;
			}
			else if(parameterMatcher.group(groupIndexOf[ParameterSource.QueryInfo.ordinal()]) != null){
				int groupIndex =  groupIndexOf[ParameterSource.QueryInfo.ordinal()];
				assert parameterMatcher.group(groupIndex).equals("query");
				++groupIndex;
				
				int index;
				if(parameterMatcher.group(groupIndex).equals("id")){
					index = 0;
				}
				else{
					assert parameterMatcher.group(groupIndex).equals("round");
					index = 1;
				}
				
				this.parameterMappingSource[parameterIndex] = ParameterSource.QueryInfo;
				this.parameterMappingLocalIndex[parameterIndex] = index;
				++groupIndex;
			}
			else{
				throw new Error();
			}
		}
		
		this.eval_statement = sql_execution_statement;		
		this.updateStatement = this.sql_translation.sqlUpdateQueryMetaEntry(this.program.meta_schema, this.id, this.numberOfOutputMinTimes, this.numberOfInputMinTimes, this.numberOfReferencedClocks);
		
		this.isMonotonic = isMonotonic;		
		
		this.program.register(this);
		
		this.usePreparedStatements = usePreparedStatements;
		this.init();
	}
	
	protected void init(){
		this.executionLock.lock();
		try{
			int tryCount = 2;
			
			while(true){
				Connection connection = this.program.bufferStore.connection();
				try
				{
					this.sql_eval_statement = this.sql_translation.prepareStatement(connection, this.eval_statement, this.usePreparedStatements);
					this.MetaEntryUpdate = this.sql_translation.prepareStatement(connection, this.updateStatement, this.usePreparedStatements);
					this.areSetEvalParameters = false;
					this.areSetMetaParameters = false;
					
					if(this.logger().isDebugEnabled()){
						if(this.loggingStatement != null){
							try{this.loggingStatement.close();} catch (SQLException e){}
						}
						this.loggingStatement = connection.createStatement();
					}
					break;
				} catch (SQLException e)
				{
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						this.debug("QUERY " + this.id + " - Connection failure. Trying again.", e);
						this.program.bufferStore.reconnect(connection);
					}
					else if(SQL_EXCEPTIONS.isMonetDBBug(e) & tryCount > 0){
						this.warn("QUERY " + this.id + " - MonetDB bug. Trying to reconnect.", e);
						this.program.bufferStore.reconnect(connection);
						--tryCount;
					}
					else{
						this.error("QUERY " + this.id, e);
						throw new RuntimeException(e);
						//TODO better exception
					}
				}
			}
		}
		finally{
			this.executionLock.unlock();
		}
	}

	@Override
	public FixedArrayNavigableMap<String, SQL_StreamBuffer> inputs()
	{
		return this.inputs;
	}

	@Override
	public FixedArray<SQL_StreamBuffer> outputs()
	{
		return new FixedArray<SQL_StreamBuffer>(this.output);
	}
	
	protected void updateClockMinTimeIncreases(){
		this.executionLock.lock();
		try{
			for(int clock = 0; clock < this.numberOfReferencedClocks; ++clock){
				this.clockMinTimes[clock] = this.referencedClocks.values().get(clock).min();	
			}
			this.minTimeClockIncreasesFunction.apply(this.clockMinTimes, 0, this.clockMinTimeIncreases, 0);
		}
		finally{
			this.executionLock.unlock();
		}
	}

	@Override
	public void execute()
	{
		this.executionLock.lock();
		try{
			
			long currentRound = this.lastRound + 1;
			
			this.info("QUERY {} (execution {}) - Start", this.id, currentRound);			
				
				long[] currentInputMinTimes;
				long[] currentOutputMinTimes;
				long[] currentOutputMinWriteTimes;
				long tupleNumber;
	
				currentInputMinTimes = this.nextInputMinTimes;
				currentOutputMinTimes = this.nextOutputMinTimes;
				currentOutputMinWriteTimes = this.nextOutputMinWriteTimes;
	
				boolean mightComputeNewTuples = false;
				int startIndex = 0;
				for(int inputIndex = 0; inputIndex < this.inputs.size(); ++inputIndex){
					SQL_StreamBuffer input = this.inputs.values().get(inputIndex);
					synchronized (input) {
						System.arraycopy(input.minWriteTimes(), 0, currentInputMinTimes, startIndex, input.numberOfMinTimes);//update inputMinTimes
						long inputTupleCount = input.tupleCount(); 
						if(inputTupleCount > this.lastTupleCounts[inputIndex]){//test if input contains any new tuple compared with the last execution of the query
							this.lastTupleCounts[inputIndex] = inputTupleCount;
							mightComputeNewTuples = true;
						}
					}
					startIndex += input.numberOfMinTimes;
				}
				assert startIndex == this.numberOfInputMinTimes;
				
				if(!this.isMonotonic){
					mightComputeNewTuples = true;
					//FIXME
				}
				if(this.type == DTSA_QueryType.INPUT){
					mightComputeNewTuples = true;
					//FIXME					
				}
				//compute currentOutputMinTimes for query
				this.minTimesFunction.apply(currentInputMinTimes, 0, currentOutputMinTimes, 0);
				
				
				{//update query info
					this.queryInfo[1] = currentRound;
				}
				
				if(this.logger().isDebugEnabled()){
					this.debug("QUERY {} (execution {})\n\tLastInputMinTimes: {}\n\tCurrentInputMinTimes: {}\n\tCurrentOutputMinTimes: {}", this.id, Long.valueOf(currentRound), this.lastInputMinTimes, currentInputMinTimes, currentOutputMinTimes);
				}
				if(this.logger().isTraceEnabled()){
					
					for(SQL_StreamBuffer input: this.inputs().values()){
						while(true){
							try{
								ResultSet tuples = this.loggingStatement.executeQuery(this.sql_translation.sqlReadBufferContent(input.schema, input.table_name));
								String tuplesTable = SQL_QueryBuildHelper.toString(tuples, -1);
								this.trace("QUERY {} (execution {}) - Snapshot of table \"{}\".\"{}\" before the execution of the eval statement:\n{}", this.id, currentRound, input.schema, input.table_name, tuplesTable);
								break;
							}
							catch(SQLException e){
								this.trace("QUERY " + this.id + " (execution " + currentRound + ") - SQLSTATE: {} - MESSAGE: {}", e.getSQLState(), e.getMessage());
								if(SQL_EXCEPTIONS.isConnectionFailure(e)){
									this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Connection failure. Trying again.", e);
									this.program.bufferStore.reconnect(this.sql_eval_statement.getConnection());
									this.init();
								}
								else if(SQL_EXCEPTIONS.isRollback(e)){
									this.debug("QUERY " + this.id + " (execution " + currentRound + ") - SQL debug table content statement rolled back. Trying again.", e);
								}
								else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
									this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Handle of prepared SQL eval statement is invalid. Repreparing statement and trying again.", e);
									this.init();
									//TODO only init the affected statement?
								}
								else{
									this.chrashed = true;
									throw e;
								}
							}				
						}
					}
					
				}
				
				if(mightComputeNewTuples){
					
					
					{//update clocks
						for(int clock = 0; clock < this.numberOfReferencedClocks; ++clock){
							this.clockTimes[clock] = this.referencedClocks.values().get(clock).now();
						}								
					}
					this.debug("QUERY {} (execution {}) - Executing eval statment.", this.id, currentRound);
					
					while(true){
						try{
							if(!this.areSetEvalParameters){
								long[] parameterSource;
								for(int parameter = 0; parameter < this.numberOfParameters; ++parameter){
			
									switch(this.parameterMappingSource[parameter]){
										case InputLastMinTimes:
											parameterSource = this.lastInputMinTimes;
											break;
										case InputNowMinTimes:
											parameterSource = currentInputMinTimes;
											break;
										case OutputLastMinTimes:
											parameterSource = this.lastOutputMinTimes;
											break;
										case OutputNowMinTimes:
											parameterSource = currentOutputMinTimes;
											break;
										case ClockNowTimes:
											parameterSource = this.clockTimes;
											break;
										case QueryInfo:
											parameterSource = this.queryInfo;
											break;
										default:
											throw new Error();
									}
									
									this.sql_eval_statement.setLong(parameter + 1, parameterSource[this.parameterMappingLocalIndex[parameter]]);
								}
								this.areSetEvalParameters = true;
							}
							
							tupleNumber = this.sql_eval_statement.executeUpdate();
							this.areSetEvalParameters = false;
							
							assert tupleNumber >= 0;
							break;
						}
						catch(SQLException e){
							this.trace("QUERY " + this.id + " (execution " + currentRound + ") - SQLSTATE: {} - MESSAGE: {}", e.getSQLState(), e.getMessage());
							if(SQL_EXCEPTIONS.isConnectionFailure(e)){
								this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Connection failure. Trying again.", e);
								this.program.bufferStore.reconnect(this.sql_eval_statement.getConnection());
								this.init();
							}
							else if(SQL_EXCEPTIONS.isRollback(e)){
								this.debug("QUERY " + this.id + " (execution " + currentRound + ") - SQL eval statement rolled back. Trying again.", e);
							}
							else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
								this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Handle of prepared SQL eval statement is invalid. Repreparing statement and trying again.", e);
								this.init();
								//TODO only init the affected statement?
							}
							else{
								this.chrashed = true;
								throw e;
							}
						}				
					}
					
					this.debug("QUERY {} (execution {}) - Executed eval statment.", this.id, currentRound);
					try{ this.sql_eval_statement.clearParameters(); } catch(SQLException e){}
					
					if(this.logger().isDebugEnabled()){
						this.debug("QUERY {} (execution {}) - Evaluation inserted {} tuples into table \"{}\".\"{}\".", this.id, currentRound, tupleNumber, this.output.schema, this.output.table_name);
						
						while(true){
							try{
								ResultSet newTuples = this.loggingStatement.executeQuery(this.sql_translation.sqlReadBufferContent(this.output.schema, this.output.table_name, this.id_attribute, this.round_attribute, this.id, currentRound));
								String newTuplesTable = SQL_QueryBuildHelper.toString(newTuples, -1);
								this.debug("QUERY {} (execution {}) - Evaluation inserted the following tuples into table \"{}\".\"{}\":\n{}", this.id, currentRound, this.output.schema, this.output.table_name, newTuplesTable);
								break;
							}
							catch(SQLException e){
								this.trace("QUERY " + this.id + " (execution " + currentRound + ") - SQLSTATE: {} - MESSAGE: {}", e.getSQLState(), e.getMessage());
								if(SQL_EXCEPTIONS.isConnectionFailure(e)){
									this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Connection failure. Trying again.", e);
									this.program.bufferStore.reconnect(this.sql_eval_statement.getConnection());
									this.init();
								}
								else if(SQL_EXCEPTIONS.isRollback(e)){
									this.debug("QUERY " + this.id + " (execution " + currentRound + ") - SQL debug table content statement rolled back. Trying again.", e);
								}
								else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
									this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Handle of prepared SQL eval statement is invalid. Repreparing statement and trying again.", e);
									this.init();
									//TODO only init the affected statement?
								}
								else{
									this.chrashed = true;
									throw e;
								}
							}				
						}
						
					}
					
				}
				else{
					this.debug("QUERY {} (execution {}) - Execution of eval statement is omited because no new tuples could have been produced.", this.id, currentRound);
					tupleNumber = 0;
				}
				
				if(this.logger().isTraceEnabled()){
					
					while(true){
						try{
							ResultSet tuples = this.loggingStatement.executeQuery(this.sql_translation.sqlReadBufferContent(this.output.schema, this.output.table_name));
							String tuplesTable = SQL_QueryBuildHelper.toString(tuples, -1);
							this.trace("QUERY {} (execution {}) - Snapshot of table \"{}\".\"{}\" after the execution of the eval statement:\n{}", this.id, currentRound, this.output.schema, this.output.table_name, tuplesTable);
							break;
						}
						catch(SQLException e){
							this.trace("QUERY " + this.id + " (execution " + currentRound + ") - SQLSTATE: {} - MESSAGE: {}", e.getSQLState(), e.getMessage());
							if(SQL_EXCEPTIONS.isConnectionFailure(e)){
								this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Connection failure. Trying again.", e);
								this.program.bufferStore.reconnect(this.sql_eval_statement.getConnection());
								this.init();
							}
							else if(SQL_EXCEPTIONS.isRollback(e)){
								this.debug("QUERY " + this.id + " (execution " + currentRound + ") - SQL debug table content statement rolled back. Trying again.", e);
							}
							else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
								this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Handle of prepared SQL eval statement is invalid. Repreparing statement and trying again.", e);
								this.init();
								//TODO only init the affected statement?
							}
							else{
								this.chrashed = true;
								throw e;
							}
						}				
					}
					
				}
				
				//access to new outputMinTimes is still protected by execution lock 
				assert this.executionLock.isHeldByCurrentThread();
				
				this.updateOutputMinWriteTimes(currentOutputMinTimes, currentOutputMinWriteTimes);
				
				this.debug("QUERY {} (execution {})\n\tOutputMinWriteTimes: {}", this.id, currentRound, currentOutputMinWriteTimes);
				
				this.debug("QUERY {} (execution {}) - Updating meta entry.", this.id, currentRound);
				int updateCount;
				while(true){
					try{
						this.sql_translation.sqlSetUpdateQueryMetaEntry(this.MetaEntryUpdate, currentRound, this.lastOutputMinWriteTimes, currentInputMinTimes);
						this.areSetMetaParameters = true;
						
						updateCount = this.MetaEntryUpdate.executeUpdate();
						this.areSetMetaParameters = false;
						
						assert updateCount >= 0;
						break;
					}
					catch(SQLException e){
						this.trace("QUERY " + this.id + " (execution " + currentRound + ") - SQLSTATE: {} - MESSAGE: {}", e.getSQLState(), e.getMessage());
						if(SQL_EXCEPTIONS.isConnectionFailure(e)){
							this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Connection failure. Trying again.", e);
							this.program.bufferStore.reconnect(this.MetaEntryUpdate.getConnection());
							this.init();
						}
						else if(SQL_EXCEPTIONS.isRollback(e)){
							this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Meta entry update rolled back. Trying again.", e);
						}
						else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
							this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Handle of prepared meta entry update statement is invalid. Repreparing statement and trying again.", e);
							this.init();
							//TODO only init the affected statement?
						}
						else{
							this.chrashed = true;
							throw e;
						}
					}				
				}
				
				this.debug("QUERY {} (execution {}) - Updated meta entry.", this.id, currentRound);
				try{ this.MetaEntryUpdate.clearParameters(); } catch(SQLException e){}
				
				synchronized (this.minWriteTimesLock) {
					this.lastRound = currentRound;
					this.nextInputMinTimes = this.lastInputMinTimes;
					this.lastInputMinTimes = currentInputMinTimes;
					
					this.nextOutputMinTimes = this.lastOutputMinTimes;
					this.lastOutputMinTimes = currentOutputMinTimes;
				
					this.nextOutputMinWriteTimes = this.lastOutputMinWriteTimes;
					this.lastOutputMinWriteTimes = currentOutputMinWriteTimes;
				}
						
				this.output.notifyWrite(tupleNumber);			
				for(SQL_StreamBuffer input: this.inputs.values()){
					input.notifyRead();
				}
				
				//TODO better log message
				this.info("QUERY {} (execution {}) - Finished", this.id, currentRound);			
				this.progLogger.info("QUERY {} (execution {}) - Finished", this.id, currentRound);

		}
		catch (SQLException e)
		{			
			//FIXME throw exeption out of query and catch there? ExecutionException?
			//TODO more differentiated reaction
			//TODO better log message
			this.error("QUERY " + this.id, e);
			this.chrashed = true;
			throw new RuntimeException(e);
		}
		catch(RuntimeException e){
			this.error("QUERY " + this.id, e);
			this.chrashed = true;
			throw e;
		}
		catch(Error e){
			this.error("QUERY " + this.id, e);
			this.chrashed = true;
			throw e;
		}
		finally{//FIXME executes finally also when catch was sucessful?
			this.executionLock.unlock();
		}
	}
	
	protected void updateOutputMinWriteTimes(final long[] outputMinTimes, final long[] outputMinWriteTimes){
		this.executionLock.lock();
		try{//only perform update if there is at least one clock that increses the min times (same for buffer -> only if one input query ... else only when input query has send notify about execution)
			this.updateClockMinTimeIncreases();
			
			long outputMinTime;
			long outputMinTimeClockIncrease;
			for(int outputMinTimeIndex = 0; outputMinTimeIndex < this.numberOfOutputMinTimes; ++outputMinTimeIndex){
				outputMinTime = outputMinTimes[outputMinTimeIndex];
				outputMinTimeClockIncrease = this.clockMinTimeIncreases[outputMinTimeIndex];
				outputMinWriteTimes[outputMinTimeIndex] = (outputMinTime < outputMinTimeClockIncrease ? outputMinTimeClockIncrease: outputMinTime);
			}
		}
		finally{
			this.executionLock.unlock();
		}
	}
	
	protected void updateOutputMinWriteTimes(){
		this.executionLock.lock();
		try{			
			synchronized (this.minWriteTimesLock) {
				this.updateOutputMinWriteTimes(this.lastOutputMinTimes, this.lastOutputMinWriteTimes);
			}
		}
		finally{
			this.executionLock.unlock();
		}
	}

	/**
	 * @return
	 */
	protected long[] outputMinWriteTimes()
	{
		boolean locked = this.executionLock.tryLock();
		try{
			if(locked & !this.chrashed){//currently not executing
				this.updateOutputMinWriteTimes();
			}
			long[] outputMinWriteTimes = new long[this.numberOfOutputMinTimes];
			synchronized (this.minWriteTimesLock) {
				System.arraycopy(this.lastOutputMinWriteTimes, 0, outputMinWriteTimes, 0, this.numberOfOutputMinTimes);	
			}
			return outputMinWriteTimes;
		}
		finally{
			if(locked){
				this.executionLock.unlock();
			}
		}
	}
	
	public synchronized long[] inputMinReadTimes()
	{
		throw new Error("Not implemented");
	}

}
